# -*- coding: utf-8 -*-

from collective.contact.plonegroup.browser.settings import invalidate_soev_cache
from collective.contact.plonegroup.browser.settings import invalidate_ssoev_cache
from collective.contact.plonegroup.config import get_registry_functions
from collective.contact.plonegroup.config import get_registry_organizations
from collective.contact.plonegroup.config import set_registry_functions
from collective.contact.plonegroup.config import set_registry_organizations
from collective.contact.plonegroup.utils import get_all_suffixes
from collective.contact.plonegroup.utils import get_organizations
from collective.contact.plonegroup.utils import get_own_organization
from collective.contact.plonegroup.utils import get_plone_group
from collective.iconifiedcategory import CAT_SEPARATOR
from copy import deepcopy
from ftw.labels.interfaces import ILabelJar
from imio.helpers.content import uuidToObject
from imio.helpers.content import validate_fields
from imio.helpers.security import generate_password
from imio.helpers.security import is_develop_environment
from plone import api
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.file import NamedImage
from Products.CMFPlone.interfaces.constrains import IConstrainTypes
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting import logger
from Products.PloneMeeting.config import MEETING_GROUP_SUFFIXES
from Products.PloneMeeting.config import MEETINGMANAGERS_GROUP_SUFFIX
from Products.PloneMeeting.config import PloneMeetingError
from Products.PloneMeeting.config import PROJECTNAME
from Products.PloneMeeting.config import registerClasses
from Products.PloneMeeting.config import TOOL_FOLDER_ANNEX_TYPES
from Products.PloneMeeting.config import TOOL_FOLDER_CATEGORIES
from Products.PloneMeeting.config import TOOL_FOLDER_CLASSIFIERS
from Products.PloneMeeting.config import TOOL_FOLDER_ITEM_TEMPLATES
from Products.PloneMeeting.config import TOOL_FOLDER_MEETING_CATEGORIES
from Products.PloneMeeting.config import TOOL_FOLDER_POD_TEMPLATES
from Products.PloneMeeting.config import TOOL_FOLDER_RECURRING_ITEMS
from Products.PloneMeeting.Extensions.imports import import_contacts
from Products.PloneMeeting.profiles import DEFAULT_USER_PASSWORD
from Products.PloneMeeting.utils import org_id_to_uid
from Products.PloneMeeting.utils import updateCollectionCriterion
from z3c.relationfield.relation import RelationValue
from zope.component import getUtility
from zope.globalrequest import getRequest
from zope.i18n import translate
from zope.intid.interfaces import IIntIds

import os
import transaction


# PloneMeeting-Error related constants -----------------------------------------
MEETING_CONFIG_ERROR = 'A validation error occurred while instantiating ' \
                       'meeting configuration "%s" with id "%s". %s'
MEETINGCONFIG_BADREQUEST_ERROR = 'There was an error during creation of ' \
    'MeetingConfig with id "%s". Original error : "%s"'


def update_labels_jar(jar, values):
    """ """
    for value in values:
        jar.add(title=value['title'],
                color=value['color'],
                by_user=value['by_user'])


class ToolInitializer:
    '''Initializes the PloneMeeting tool based on information from a given
       PloneMeeting profile.'''
    successMessage = "The PloneMeeting tool has been successfully initialized."
    noDataMessage = "No data to import for this profile"

    def __init__(self, context, productname):
        self.profilePath = context._profile_path
        # productname default's name space is 'Products'.
        # If a name space is found, then Products namespace is not used
        self.productname = '.' in productname and productname or 'Products.%s' % productname
        self.request = getRequest()
        self.portal = context.getSite()
        self.tool = self.portal.portal_plonemeeting
        # set correct title
        self.tool.setTitle(translate('pm_configuration',
                           domain='PloneMeeting',
                           context=self.request))
        self.profileData = self.getProfileData()

    def _init_tool(self, force=False):
        # Initialize the tool if we have data
        if not self.profileData:
            return
        # initialize the tool and configure the contacts directory
        # only if it was not already done before
        # by another profile, it is the case if some MeetingConfigs exist
        if force or not self.tool.objectIds('MeetingConfig'):
            for k, v in self.profileData.getData().iteritems():
                if k == 'advisersConfig':
                    # turn org id into uid
                    for adviserConfig in v:
                        adviserConfig['org_uids'] = [
                            org_id_to_uid(org_id) for org_id in adviserConfig['org_uids']]
                exec 'self.tool.set%s%s(v)' % (k[0].upper(), k[1:])
            # contacts directory
            if self.profileData.directory_position_types:
                self.portal.contacts.position_types = self.profileData.directory_position_types
            # contacts DashboardPODTemplate
            for descr in self.profileData.contactsTemplates:
                if descr.id not in self.portal.contacts:
                    self.addPodTemplate(self.portal.contacts, descr, source=self.profilePath)
            # this will especially apply advisersConfig
            self.tool.at_post_edit_script()

    def getProfileData(self):
        '''Loads, from the current profile, the data to import into the tool:
           meeting config(s), categories, etc.'''
        pp = self.profilePath
        if not pp:
            return
        profileModule = pp[pp.rfind(self.productname.replace('.', '/')):].replace('/', '.')
        profileModule = profileModule.replace('\\', '.')
        data = ''
        module_path = 'from %s.import_data import data' % profileModule
        exec module_path
        return data

    def run(self):
        self.data = self.profileData
        if not self.data:
            return self.noDataMessage
        # Register classes again, after model adaptations have been performed
        # (see comment in __init__.py)
        registerClasses()
        # if we already have existing organizations, we do not add additional ones
        own_org = get_own_organization()
        alreadyHaveGroups = bool(own_org.objectValues())
        savedMeetingConfigsToCloneTo = {}
        savedOrgsData = {}
        if not alreadyHaveGroups or self.data.forceAddUsersAndGroups:
            # 1) create organizations so we have org UIDS to initialize 'fct_orgs'
            orgs, active_orgs, savedOrgsData = self.addOrgs(self.data.orgs)
            # 2) create plonegroup functions (suffixes) to create Plone groups
            functions = get_registry_functions()
            function_ids = [function['fct_id'] for function in functions]
            # append new functions
            # late import for multi config profile (MeetingLalouviere)
            from Products.PloneMeeting.config import EXTRA_GROUP_SUFFIXES
            suffixes = MEETING_GROUP_SUFFIXES + EXTRA_GROUP_SUFFIXES
            for suffix in suffixes:
                if suffix['fct_id'] not in function_ids:
                    copied_suffix = suffix.copy()
                    copied_suffix['fct_title'] = translate(suffix['fct_title'],
                                                           domain='PloneMeeting',
                                                           context=self.request)
                    # if org_path not found, do not fail but log, it is often the case in tests
                    # in which we do not add additional organizations because it breaks some tests
                    copied_suffix['fct_orgs'] = []
                    for org_path in suffix['fct_orgs']:
                        try:
                            fct_org = own_org.restrictedTraverse(org_path)
                        except KeyError:
                            logger.warning(
                                "Could not find an organization with path {0} "
                                "while setting 'fct_orgs' for {1}".format(
                                    org_path, suffix['fct_id']))
                            continue
                        copied_suffix['fct_orgs'].append(fct_org.UID())
                    functions.append(copied_suffix)
            # 3) manage organizations, set every organizations so every Plone groups are created
            # then disable orgs that are not active
            invalidate_soev_cache()
            invalidate_ssoev_cache()
            already_active_orgs = get_registry_organizations()
            org_uids = [org_uid for org_uid in get_organizations(only_selected=False, the_objects=False)
                        if org_uid not in already_active_orgs]
            set_registry_organizations(org_uids)
            set_registry_functions(functions)
            active_org_uids = [org.UID() for org in active_orgs]
            set_registry_organizations(already_active_orgs + active_org_uids)
            # 4) add users to Plone groups
            self.addUsers(self.data.orgs)
            # 5) now that organizations are created, we add persons and held_positions
            self.addPersonsAndHeldPositions(self.data.persons, source=self.profilePath)

        # initialize the tool when organizations have been created
        self._init_tool()

        created_cfgs = []
        for mConfig in self.data.meetingConfigs:
            # XXX we need to defer the management of the 'meetingConfigsToCloneTo'
            # defined on the mConfig after the creation of every mConfigs because
            # if we defined in mConfig1.meetingConfigsToCloneTo the mConfig2 id,
            # it will try to getattr this meetingConfig2 id that does not exist yet...
            # so save defined values, removed them from mConfig and manage that after
            savedMeetingConfigsToCloneTo[mConfig.id] = mConfig.meetingConfigsToCloneTo
            mConfig.meetingConfigsToCloneTo = []
            cfg = self.createMeetingConfig(mConfig, source=self.profilePath)
            if cfg:
                created_cfgs.append(cfg)
                self._finishConfigFor(cfg, data=mConfig)

        # manage other_mc_correspondences
        for created_cfg in created_cfgs:
            self._manageOtherMCCorrespondences(created_cfg)
            self._manage_after_scan_change_annex_type_to(created_cfg)

        # now that every meetingConfigs have been created, we can manage the meetingConfigsToCloneTo
        # and orgs advice states related fields
        for mConfigId in savedMeetingConfigsToCloneTo:
            if not savedMeetingConfigsToCloneTo[mConfigId]:
                continue
            # initialize the attribute on the meetingConfig and call _updateCloneToOtherMCActions
            cfg = getattr(self.tool, mConfigId)
            # validate the MeetingConfig.meetingConfigsToCloneTo data that we are about to set
            # first replace cfg1 and cfg2 by corresponding cfg id
            adapted_cfgsToCloneTo = deepcopy(savedMeetingConfigsToCloneTo[mConfigId])
            for cfgToCloneTo in adapted_cfgsToCloneTo:
                cfgToCloneTo['meeting_config'] = self.cfg_num_to_id(cfgToCloneTo['meeting_config'])
            error = cfg.validate_meetingConfigsToCloneTo(adapted_cfgsToCloneTo)
            if error:
                raise PloneMeetingError(MEETING_CONFIG_ERROR % (cfg.Title(), cfg.getId(), error))
            cfg.setMeetingConfigsToCloneTo(adapted_cfgsToCloneTo)
            cfg._updateCloneToOtherMCActions()
        for org_uid, values in savedOrgsData.items():
            org = uuidToObject(org_uid, unrestricted=True)
            # turn cfg1__state__itemcreated into meeting-config-id__state__itemcreated
            org.item_advice_states = self._correct_advice_states(values['item_advice_states'])
            org.item_advice_edit_states = self._correct_advice_states(values['item_advice_edit_states'])
            org.item_advice_view_states = self._correct_advice_states(values['item_advice_view_states'])
            org.groups_in_charge = [org_id_to_uid(group_id) for group_id in values['groups_in_charge']]

        # finally, create the current user (admin) member area
        self.portal.portal_membership.createMemberArea()
        # at the end, add users outside PloneMeeting groups because
        # they could have to be added in groups created by the MeetingConfig
        if not alreadyHaveGroups:
            # adapt userDescr.ploneGroups to turn cfg_num into cfg_id
            self.addUsersOutsideGroups(self.data.usersOutsideGroups)

        # commit before continuing so elements like scales on annex types are correctly saved
        transaction.commit()
        return self.successMessage

    def _correct_advice_states(self, advice_states):
        """ """
        return ['{0}__state__{1}'.format(
                self.cfg_num_to_id(v.split('__state__')[0]),
                v.split('__state__')[1]) for v in advice_states]

    def _finishConfigFor(self, cfg, data):
        """When the MeetingConfig has been created, some parameters still need to be applied
           because they need the MeetingConfig to exist."""
        # apply the meetingTopicStates to the 'searchnotdecidedmeetings' DashboardCollection
        updateCollectionCriterion(cfg.searches.searches_meetings.searchnotdecidedmeetings,
                                  'review_state',
                                  list(data.meetingTopicStates))
        # apply the maxDaysDecisions to the 'searchlastdecisions' DashboardCollection
        updateCollectionCriterion(cfg.searches.searches_decisions.searchlastdecisions,
                                  'meeting_date',
                                  unicode(data.maxDaysDecisions))
        # apply the decisionTopicStates to the 'searchlastdecisions'
        # and 'searchalldecision' DashboardCollections
        updateCollectionCriterion(cfg.searches.searches_decisions.searchlastdecisions,
                                  'review_state',
                                  list(data.decisionTopicStates))
        # select correct default view
        meetingAppDefaultView = data.meetingAppDefaultView
        if meetingAppDefaultView in cfg.searches.searches_items.objectIds():
            cfg._set_default_faceted_search(meetingAppDefaultView)
        else:
            error = 'meetingAppDefaultView : No DashboardCollection with id %s' % meetingAppDefaultView
            raise PloneMeetingError(MEETING_CONFIG_ERROR % (cfg.Title(), cfg.getId(), error))

        # now we can set values for dashboard...Filters fields as the 'searches' folder has been created
        for fieldName in ('dashboardItemsListingsFilters',
                          'dashboardMeetingAvailableItemsFilters',
                          'dashboardMeetingLinkedItemsFilters'):
            field = cfg.getField(fieldName)
            # we want to validate the vocabulay, as if enforceVocabulary was True
            error = field.validate_vocabulary(cfg, cfg.getField(field.getName()).get(cfg), {})
            if error:
                raise PloneMeetingError(MEETING_CONFIG_ERROR % (cfg.Title(), cfg.getId(), error))

        if data.addContactsCSV:
            output = import_contacts(self.portal, path=self.profilePath)
            logger.info(output)
            selectableOrderedContacts = cfg.getField('orderedContacts').Vocabulary(cfg).keys()
            cfg.setOrderedContacts(selectableOrderedContacts)

        # turn contact path to uid
        for org_storing_field in ('orderedContacts', ):
            org_storing_data = getattr(data, org_storing_field, [])
            if org_storing_data:
                contact_uids = []
                for contact_path in org_storing_data:
                    try:
                        contact_uid = self.portal.contacts.restrictedTraverse(contact_path).UID()
                        contact_uids.append(contact_uid)
                    except KeyError:
                        logger.warning('While computing "{0}", could not get contact at "{1}"'.format(
                            org_storing_field, contact_path))
                cfg.getField(org_storing_field).set(cfg, contact_uids)

        # set default labels
        if data.defaultLabels:
            jar = ILabelJar(cfg)
            update_labels_jar(jar, values=data.defaultLabels)

        # disable relevant dashboard collections
        for collection_path in data.disabled_collections:
            try:
                collection = cfg.searches.restrictedTraverse(collection_path)
            except AttributeError:
                logger.warning('While disabling collections, no collection found at {0}, ...'.format(
                    collection_path))
            collection.enabled = False
            collection.reindexObject(idxs=['enabled'])

    def _manageOtherMCCorrespondences(self, cfg):
        def _convert_to_real_other_mc_correspondences(annex_type):
            """ """
            real_other_mc_correspondences = []
            # we have a content_category id prefixed with cfg id
            # like meeting-config-test_-_annexes_types_-_item_annexes_-_annex
            # but we need the UID of the corresponding annexType
            for other_mc_correspondence in annex_type.other_mc_correspondences:
                steps = other_mc_correspondence.split(CAT_SEPARATOR)
                cfg_id = self.cfg_num_to_id(steps[0])
                other_cfg = self.tool.get(cfg_id)
                corresponding_annex_type = other_cfg
                for step in steps[1:]:
                    corresponding_annex_type = corresponding_annex_type[step]
                real_other_mc_correspondences.append(corresponding_annex_type.UID())
            annex_type.other_mc_correspondences = set(real_other_mc_correspondences)

        # finish configuration of annexType.other_mc_correspondences
        # for ItemAnnexContentCategory and ItemAnnexContentSubcategory
        for annex_group in cfg.annexes_types.objectValues():
            if 'ItemAnnexContentCategory' in IConstrainTypes(annex_group).getLocallyAllowedTypes():
                for annex_type in annex_group.objectValues():
                    if annex_type.other_mc_correspondences:
                        _convert_to_real_other_mc_correspondences(annex_type)
                        for subType in annex_type.objectValues():
                            _convert_to_real_other_mc_correspondences(subType)

    def _manage_after_scan_change_annex_type_to(self, cfg):
        """We have the new annex type path, we need it's UID."""
        def _convert_to_real_after_scan_change_annex_type_to(annex_type):
            # get the annex_type and replaced it by it's UID
            other_annex_type = cfg.annexes_types.unrestrictedTraverse(
                annex_type.after_scan_change_annex_type_to)
            annex_type.after_scan_change_annex_type_to = other_annex_type.UID()
        for annex_group in cfg.annexes_types.objectValues():
            for annex_type in annex_group.objectValues():
                if annex_type.after_scan_change_annex_type_to:
                    _convert_to_real_after_scan_change_annex_type_to(annex_type)
                    for sub_type in annex_type.objectValues():
                        if sub_type.after_scan_change_annex_type_to:
                            _convert_to_real_after_scan_change_annex_type_to(sub_type)

    def createMeetingConfig(self, configData, source):
        '''Creates a new meeting configuration from p_configData which is a
           MeetingConfigDescriptor instance. p_source is a string that
           corresponds to the absolute path of a profile; additional (binary)
           data like images or templates that need to be attached to some
           sub-objects of the meeting config will be searched there.'''
        cData = configData.getData()
        # turn org ids into org uids
        for field_name in ['orderedAssociatedOrganizations',
                           'orderedContacts',
                           'orderedGroupsInCharge',
                           'orderedItemInitiators',
                           'powerAdvisersGroups',
                           'selectableCopyGroups',
                           'selectableAdvisers',
                           'usingGroups']:
            data = cData.get(field_name)
            try:
                data = [org_id_to_uid(suffixed_group_id) for suffixed_group_id in data
                        if org_id_to_uid(suffixed_group_id, raise_on_error=False)]
            except KeyError:
                logger.warning(
                    'Error while turning org ids to org uids for field_name {0}'.format(
                        field_name))
                data = []
            cData[field_name] = data
        # manage customAdvisers, turn 'org' value from org id to uid
        ca = cData.get('customAdvisers')
        adapted_ca = []
        for v in ca:
            new_value = v.copy()
            new_value['org'] = org_id_to_uid(new_value['org'])
            adapted_ca.append(new_value)
        cData['customAdvisers'] = adapted_ca
        # return if already exists
        if cData['id'] in self.tool.objectIds():
            logger.info(
                'A MeetingConfig with id {0} already exists, passing...'.format(
                    cData['id']))
            return
        self.tool.invokeFactory('MeetingConfig', **cData)
        cfgId = configData.id
        cfg = getattr(self.tool, cfgId)
        cfg._at_creation_flag = True
        # for tests where config id is shuffled, save the real id
        if "__real_id__" in cData:
            cfg.__real_id__ = cData["__real_id__"]
        # TextArea fields are not set properly.
        for field in cfg.Schema().fields():
            fieldName = field.getName()
            widgetName = field.widget.getName()
            if (widgetName == 'TextAreaWidget') and fieldName in cData:
                field.set(cfg, cData[fieldName], mimetype='text/html')
        # call processForm passing dummy values so existing values are not touched
        cfg.processForm(values={'dummy': None})
        # Validates meeting config (validation seems not to be triggered
        # automatically when an object is created from code).
        errors = []
        for field in cfg.Schema().fields():
            error = field.validate(cfg.getField(field.getName()).get(cfg), cfg)
            if error:
                errors.append("'%s': %s" % (field.getName(), error))
        if errors:
            raise PloneMeetingError(MEETING_CONFIG_ERROR % (
                safe_unicode(cfg.Title()), cfg.getId(), u'\n'.join(errors)))

        if not configData.active:
            self.portal.portal_workflow.doActionFor(cfg, 'deactivate')
        # Adds the sub-objects within the config: categories, classifiers, items in config, ...
        for descr in configData.categories:
            self.addCategory(cfg, descr)
        for descr in configData.classifiers:
            self.addCategory(cfg, descr, folder_id=TOOL_FOLDER_CLASSIFIERS)
        for descr in configData.meetingcategories:
            self.addCategory(cfg, descr, folder_id=TOOL_FOLDER_MEETING_CATEGORIES)
        for descr in configData.recurringItems:
            self.addItemToConfig(cfg, descr)
        for descr in configData.itemTemplates:
            self.addItemToConfig(cfg, descr, isRecurring=False)
        # configure ContentCategoryGroups before adding annex_types
        # this will enable confidentiality, to_sign/signed, ... if necessary
        for category_group_id, attrs in configData.category_group_activated_attrs.items():
            category_group = cfg.annexes_types.get(category_group_id)
            for attr in attrs:
                if not hasattr(category_group, attr):
                    raise PloneMeetingError(
                        'Attribute {0} does not exist on {1} of {2}'.format(
                            attr, category_group_id, cfgId))
                setattr(category_group, attr, True)
        for descr in configData.annexTypes:
            self.addAnnexType(cfg, descr, source)
        # first create style templates so it exist before being used by a pod template
        for descr in configData.styleTemplates:
            self.addPodTemplate(cfg, descr, source)
        for descr in configData.podTemplates:
            self.addPodTemplate(cfg, descr, source)
        # manage MeetingManagers
        groupsTool = self.portal.portal_groups
        for userId in configData.meetingManagers:
            groupsTool.addPrincipalToGroup(userId, '{0}_{1}'.format(
                cfg.getId(), MEETINGMANAGERS_GROUP_SUFFIX))
        # manage annex confidentiality, enable it on relevant CategoryGroup
        if configData.itemAnnexConfidentialVisibleFor:
            cfg.annexes_types.item_annexes.confidentiality_activated = True
            cfg.annexes_types.item_decision_annexes.confidentiality_activated = True
        if configData.adviceAnnexConfidentialVisibleFor:
            cfg.annexes_types.advice_annexes.confidentiality_activated = True
        if configData.meetingAnnexConfidentialVisibleFor:
            cfg.annexes_types.meeting_annexes.confidentiality_activated = True
        return cfg

    def addCategory(self, cfg, descr, folder_id=TOOL_FOLDER_CATEGORIES):
        '''Creates a category in given p_folder_id from p_descr,
           a CategoryDescriptor instance.'''
        folder = getattr(cfg, folder_id)

        data = descr.getData()
        cat = api.content.create(container=folder,
                                 type='meetingcategory',
                                 **data)
        # adapt org related values as we have org id on descriptor and we need to set org UID
        if cat.using_groups:
            cat.using_groups = ([org_id_to_uid(using_group) for using_group in cat.using_groups])
        if cat.groups_in_charge:
            cat.groups_in_charge = ([org_id_to_uid(gic) for gic in cat.groups_in_charge])
        return cat

    def addItemToConfig(self, cfg, descr, isRecurring=True):
        '''Adds a recurring item or item template
           from a RecurringItemDescriptor or a ItemTemplateDescriptor
           depending on p_isRecurring.'''
        if isRecurring:
            folder = getattr(cfg, TOOL_FOLDER_RECURRING_ITEMS)
        else:
            folder = getattr(cfg, TOOL_FOLDER_ITEM_TEMPLATES)
        data = descr.__dict__
        itemType = isRecurring and \
            cfg.getItemTypeName(configType='MeetingItemRecurring') or \
            cfg.getItemTypeName(configType='MeetingItemTemplate')
        folder.invokeFactory(itemType, **data)
        item = getattr(folder, descr.id)
        # adapt org related values as we have org id on descriptor and we need to set org UID
        if item.proposingGroup:
            item.setProposingGroup(org_id_to_uid(item.proposingGroup))
        if item.groupsInCharge:
            item.setGroupsInCharge([org_id_to_uid(grp_in_charge)
                                    for grp_in_charge in item.groupsInCharge])
        if item.proposingGroupWithGroupInCharge:
            proposing_group_id, group_in_charge_id = item.proposingGroupWithGroupInCharge.split(
                '__groupincharge__')
            proposing_group_uid = org_id_to_uid(proposing_group_id)
            group_in_charge_uid = org_id_to_uid(group_in_charge_id)
            item.setProposingGroup(proposing_group_uid)
            item.setGroupsInCharge((group_in_charge_uid, ))
            item.proposingGroupWithGroupInCharge = '{0}__groupincharge__{1}'.format(
                proposing_group_uid, group_in_charge_uid)
        if item.associatedGroups:
            item.setAssociatedGroups(
                [org_id_to_uid(associated_group)
                 for associated_group in item.associatedGroups])
        if item.templateUsingGroups:
            item.setTemplateUsingGroups(
                [org_id_to_uid(template_using_group)
                 for template_using_group in item.templateUsingGroups])
        # disable _at_rename_after_creation for itemTemplates and recurringItems
        item._at_rename_after_creation = False
        # call processForm passing dummy values so existing values are not touched
        item.processForm(values={'dummy': None})
        return item

    def addAnnexType(self, cfg, at, source):
        '''Adds an annex type from a AnnexTypeDescriptor p_at.'''
        folder = getattr(cfg, TOOL_FOLDER_ANNEX_TYPES)
        # create (ItemAnnex)ContentCategory in right subfolder (ContentCategoryGroup)
        portal_type = 'ContentCategory'
        sub_portal_type = 'ContentSubcategory'
        if at.relatedTo == 'item':
            portal_type = 'ItemAnnexContentCategory'
            sub_portal_type = 'ItemAnnexContentSubcategory'
            categoryGroupId = 'item_annexes'
        elif at.relatedTo == 'item_decision':
            portal_type = 'ItemAnnexContentCategory'
            sub_portal_type = 'ItemAnnexContentSubcategory'
            categoryGroupId = 'item_decision_annexes'
        elif at.relatedTo == 'advice':
            categoryGroupId = 'advice_annexes'
        elif at.relatedTo == 'meeting':
            categoryGroupId = 'meeting_annexes'

        # The image must be retrieved on disk from a profile
        iconPath = '%s/images/%s' % (source, at.icon)
        data = self.find_binary(iconPath)
        contentCategoryFile = NamedBlobImage(data, filename=at.icon)
        annexType = api.content.create(
            id=at.id,
            type=portal_type,
            title=at.title,
            predefined_title=at.predefined_title,
            icon=contentCategoryFile,
            container=getattr(folder, categoryGroupId),
            to_print=at.to_print,
            confidential=at.confidential,
            to_sign=at.to_sign,
            signed=at.signed,
            enabled=at.enabled,
            show_preview=at.show_preview,
            description=at.description,
            after_scan_change_annex_type_to=at.after_scan_change_annex_type_to,
            only_pdf=at.only_pdf,
        )
        # store an empty set in other_mc_correspondences for validation
        # then store intermediate value that will be reworked at the end
        # of the MeetingConfig instanciation
        if portal_type == 'ItemAnnexContentCategory':
            annexType.other_mc_correspondences = set()
        validate_fields(annexType, raise_on_errors=True)
        if portal_type == 'ItemAnnexContentCategory':
            annexType.other_mc_correspondences = at.other_mc_correspondences
            annexType.only_for_meeting_managers = at.only_for_meeting_managers

        for subType in at.subTypes:
            annexSubType = api.content.create(
                id=subType.id,
                type=sub_portal_type,
                title=subType.title,
                predefined_title=subType.predefined_title,
                container=annexType,
                to_print=subType.to_print,
                confidential=subType.confidential,
                to_sign=subType.to_sign,
                signed=subType.signed,
                enabled=subType.enabled,
                show_preview=subType.show_preview,
                after_scan_change_annex_type_to=subType.after_scan_change_annex_type_to,
                description=subType.description,
            )
            if sub_portal_type == 'ItemAnnexContentSubcategory':
                annexSubType.other_mc_correspondences = set()
            validate_fields(annexSubType, raise_on_errors=True)
            if sub_portal_type == 'ItemAnnexContentSubcategory':
                annexSubType.other_mc_correspondences = subType.other_mc_correspondences
                annexType.only_for_meeting_managers = at.only_for_meeting_managers

        return annexType

    def addPodTemplate(self, container, pt, source):
        '''Adds a POD template from p_pt (a PodTemplateDescriptor instance).'''
        cfg = None
        if container.portal_type == 'MeetingConfig':
            cfg = container
            dest_folder = getattr(container, TOOL_FOLDER_POD_TEMPLATES)
        else:
            # contacts DashboardPODTemplate
            dest_folder = container
        # The template must be retrieved on disk from a profile or use another pod_template
        odt_file = None
        pod_template_to_use = None
        if pt.odt_file:
            # pt.odt_file may be a full path or a relative path
            filePath = pt.odt_file.startswith('/') and pt.odt_file or '%s/templates/%s' % (
                source, pt.odt_file)
            data = self.find_binary(filePath)
            odt_file = NamedBlobFile(
                data=data,
                contentType='application/vnd.oasis.opendocument.text',
                # pt.odt_file could be relative (../../other_profile/templates/sample.odt)
                filename=safe_unicode(pt.odt_file.split('/')[-1]),
            )
        elif pt.pod_template_to_use['cfg_id']:
            pod_template_to_use_cfg = self.tool.get(pt.pod_template_to_use['cfg_id'])
            if not pod_template_to_use_cfg:
                logger.warning(
                    'Cfg with id {0} not found when adding Pod template {1}, template was not added'.format(
                        pt.pod_template_to_use['cfg_id'], pt.pod_template_to_use['template_id']))
                return
            pod_template = pod_template_to_use_cfg.podtemplates.get(pt.pod_template_to_use['template_id'])
            if not pod_template:
                logger.warning(
                    'Pod template with id {0} not found in cfg with id {1}, template was not added'.format(
                        pt.pod_template_to_use['template_id'], pt.pod_template_to_use['cfg_id']))
                return
            pod_template_to_use = pod_template.UID()
        else:
            raise PloneMeetingError(
                'A PodTemplateDescriptor must have a defined odt_file or pod_template_to_use!')
        data = pt.getData(odt_file=odt_file, pod_template_to_use=pod_template_to_use)

        if data['is_style']:
            podType = 'StyleTemplate'
        else:
            podType = pt.dashboard and 'DashboardPODTemplate' or 'ConfigurablePODTemplate'

            # turn the pod_portal_types from MeetingItem to MeetingItemShortname
            adapted_pod_portal_types = []
            for pod_portal_type in data['pod_portal_types']:
                if pod_portal_type.startswith('Meeting'):
                    pod_portal_type = pod_portal_type + cfg.shortName
                adapted_pod_portal_types.append(pod_portal_type)
            data['pod_portal_types'] = adapted_pod_portal_types

            if podType == 'DashboardPODTemplate':
                # parameter use_objects is excluded by default as only relevant for DashboardPODTemplates
                data['use_objects'] = pt.use_objects
                # manage dashboard_collections from dashboard_collection_ids
                # we have ids and we need UIDs
                res = []
                if cfg:
                    collections_container = cfg.searches
                else:
                    # contacts
                    collections_container = container
                for coll_id in pt.dashboard_collections_ids:
                    for folder in collections_container.objectValues('ATFolder'):
                        if coll_id in folder.objectIds():
                            collection = getattr(folder, coll_id)
                            break
                    res.append(collection.UID())
                data['dashboard_collections'] = res
            for sub_template in data['merge_templates']:
                sub_template['template'] = dest_folder.get(sub_template['template']).UID()

        # associate style template with pod template if necessary
        if not data['is_style'] and data['style_template']:
            # we have a list of style templates
            styles_uids = []
            for style_template in data['style_template']:
                style_template_obj = dest_folder.get(style_template)
                if style_template_obj.id in data['style_template']:
                    styles_uids.append(style_template_obj.UID())

            data['style_template'] = styles_uids

        podTemplate = api.content.create(
            type=podType,
            container=dest_folder,
            **data)
        validate_fields(podTemplate, raise_on_errors=True)
        return podTemplate

    def addUsersOutsideGroups(self, usersOutsideGroups):
        '''Create users that are outside any PloneMeeting group (like WebDAV
           users or users that are in groups created by MeetingConfigs).'''
        for userDescr in usersOutsideGroups:
            self.addUser(userDescr)

    def addOrgs(self, org_descriptors, defer_data=True):
        '''Creates organizations (a list of OrgaDescriptor instances) in the contact own organization.'''
        own_org = get_own_organization()
        orgs = []
        active_orgs = []
        savedOrgsData = {}
        for org_descr in org_descriptors:
            if org_descr.parent_path:
                # find parent organization following parent path from container
                container = own_org.restrictedTraverse(org_descr.parent_path)
            else:
                container = own_org
            # Maybe the organization already exists?
            # It could be the case if we are reapplying a configuration
            if org_descr.id in container.objectIds():
                continue

            # save some informations that will be reinjected at the end
            data = org_descr.getData()
            if defer_data:
                # org is not created and we needs its uid...
                savedOrgsData['dummy'] = {'item_advice_states': data['item_advice_states'],
                                          'item_advice_edit_states': data['item_advice_edit_states'],
                                          'item_advice_view_states': data['item_advice_view_states'],
                                          'groups_in_charge': data['groups_in_charge']}
                data['item_advice_states'] = []
                data['item_advice_edit_states'] = []
                data['item_advice_view_states'] = []
                data['groups_in_charge'] = []

            org = api.content.create(container=container, type='organization', **data)

            if defer_data:
                # finalize savedOrgsData, store org uid instead 'dummy'
                savedOrgsData[org.UID()] = savedOrgsData['dummy'].copy()
                del savedOrgsData['dummy']
            validate_fields(org, raise_on_errors=True)
            orgs.append(org)
            if org_descr.active:
                active_orgs.append(org)
        return orgs, active_orgs, savedOrgsData

    def addUsers(self, org_descriptors):
        '''Creates Plone users and add it to linked Plone groups.'''
        plone_utils = self.portal.plone_utils
        # if we are in dev, we use DEFAULT_USER_PASSWORD, else we will generate a
        # password that is compliant with the current password policy...
        if is_develop_environment():
            password = DEFAULT_USER_PASSWORD
        else:
            password = generate_password()
        msg = "The password used for added users is %s" % (password or DEFAULT_USER_PASSWORD)
        logger.info(msg)
        # add a portal_message so admin adding the Plone site knows password
        plone_utils.addPortalMessage(msg, 'warning')

        member_tool = api.portal.get_tool('portal_membership')

        own_org = get_own_organization()
        plonegroup_org_uids = get_registry_organizations()
        for org_descr in org_descriptors:
            if org_descr.parent_path:
                # find parent organization following parent path from container
                container = own_org.restrictedTraverse(org_descr.parent_path)
            else:
                container = own_org
            org = container.get(org_descr.id)
            # Create users
            for userDescr in org_descr.getUsers():
                # if we defined a generated password here above, we use it
                # either we use the password provided in the applied profile
                if password:
                    userDescr.password = password
                self.addUser(userDescr)
            # Add users in the correct Plone groups.
            org_uid = org.UID()
            if org_uid in plonegroup_org_uids:
                for suffix in get_all_suffixes(org_uid):
                    plone_group = get_plone_group(org_uid, suffix)
                    group_members = plone_group.getMemberIds()
                    # protect in case we have suffixes only for some groups
                    for userDescr in getattr(org_descr, suffix, []):
                        if userDescr.id not in group_members:
                            api.group.add_user(group=plone_group, username=userDescr.id)
                            if userDescr.create_member_area:
                                member_tool.createMemberArea(userDescr.id)

    def addUser(self, userData):
        '''Adds a new Plone user from p_userData which is a UserDescriptor
           instance if it does not already exist.'''
        usersDb = self.portal.acl_users.source_users
        if usersDb.getUserById(userData.id) or userData.id == 'admin':
            return  # Already exists.
        self.portal.portal_registration.addMember(
            userData.id, userData.password,
            ['Member'] + userData.globalRoles,
            properties={'username': userData.id,
                        'email': userData.email,
                        'fullname': userData.fullname or ''})
        # Add the user to some Plone groups
        groupsTool = self.portal.portal_groups
        for groupDescr in userData.ploneGroups:
            # Create the group if it does not exist, turn cfg_num into cfg_id
            # We have something like cfg1_powerobservers
            group_data = groupDescr.getData()
            if group_data['id'].startswith('cfg'):
                cfg_id = self.cfg_num_to_id(group_data['id'][0:4])
                group_data['id'] = cfg_id + group_data['id'][4:]
            if group_data['title'].startswith('cfg'):
                cfg_id = self.cfg_num_to_id(group_data['title'][0:4])
                group_data['title'] = cfg_id + group_data['title'][4:]
            if not groupsTool.getGroupById(group_data['id']):
                groupsTool.addGroup(group_data['id'], title=group_data['title'])
                if groupDescr.roles:
                    groupsTool.setRolesForGroup(group_data['id'],
                                                groupDescr.roles)
            groupsTool.addPrincipalToGroup(userData.id, group_data['id'])

    def addPersonsAndHeldPositions(self, person_descriptors, source):
        '''Creates persons and eventual held_positions.'''
        own_org = get_own_organization()
        container = own_org.aq_inner.aq_parent
        intids = getUtility(IIntIds)
        for person_descr in person_descriptors:
            # create the person
            person = api.content.create(
                container=container,
                type='person',
                **person_descr.getData())
            # person.photo is the name of the photo to use stored in /images
            if person.photo:
                photo_path = '%s/images/%s' % (source, person.photo)
                data = self.find_binary(photo_path)
                photo_file = NamedImage(data, filename=person.photo)
                person.photo = photo_file
                validate_fields(person, raise_on_errors=True)
            for held_pos_descr in person_descr.held_positions:
                # get the position
                data = held_pos_descr.getData()
                org = container.restrictedTraverse(data['position'])
                data['position'] = RelationValue(intids.getId(org))
                held_position = api.content.create(
                    container=person,
                    type='held_position',
                    **data)
                validate_fields(held_position, raise_on_errors=True)

    def cfg_num_to_id(self, cfg_num):
        """ """
        # make sure we have something like cfg1, cfg2, ...
        # if not, then cfg_num is the cfg_id
        if cfg_num.startswith('cfg'):
            num = int(cfg_num[3:]) - 1
            cfg_descr = self.data.meetingConfigs[num]
            return cfg_descr.id
        else:
            return cfg_num

    def find_binary(self, path):
        """If a binary (odt, icon, photo, ...) is not found in current profile,
           try to get it from Products.PloneMeeting.profiles.testing."""
        try:
            f = file(path, 'rb')
        except IOError:
            logger.warn(
                'Could not get binary at "{0}", trying to get it from the testing profile...'.format(path))
            import Products.PloneMeeting

            # get binary folder, last part of path (templates, images, ...)
            splitted_path = path.split('/')
            file_folder, filename = splitted_path[-2], splitted_path[-1]
            pm_path = os.path.join(
                os.path.dirname(Products.PloneMeeting.__file__),
                'profiles/testing',
                file_folder,
                filename)
            f = file(pm_path, 'rb')
        data = f.read()
        f.close()
        return data


def isTestOrArchiveProfile(context):
    isTest = context.readDataFile("PloneMeeting_testing_marker.txt")
    isArchive = context.readDataFile("PloneMeeting_archive_marker.txt")
    return isTest or isArchive


def initializeTool(context):
    '''Initialises the PloneMeeting tool based on information from the current
       profile.'''
    # This method is called by several profiles: testing, archive. Because of a bug
    # in portal_setup, the method can be wrongly called by the default profile.
    if not isTestOrArchiveProfile(context):
        return
    # Installs PloneMeeting if not already done
    pqi = context.getSite().portal_quickinstaller
    # Now that we do not run this profile from elsewhere than portal_setup
    # We had to install PloneMeeting first...
    # pqi.listInstalledProducts()
    if not pqi.isProductInstalled('PloneMeeting'):
        profileId = u'profile-Products.PloneMeeting:default'
        context.getSite().portal_setup.runAllImportStepsFromProfile(profileId)
    # Initialises data from the profile.
    return ToolInitializer(context, PROJECTNAME).run()



# -*- coding: utf-8 -*-

from ftw.labels.interfaces import ILabelJar
from persistent.mapping import PersistentMapping
from plone.app.textfield.value import RichTextValue
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from zope.i18n import translate


class Migrate_To_4102(Migrator):

    def _updateFTWLabelsStorage(self):
        """ftw.labels jar was created using dict we need PersistentMappings..."""
        logger.info("Updating ftw.labels jar for every MeetingConfigs...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            jar_storage = ILabelJar(cfg).storage
            for k, v in jar_storage.items():
                jar_storage[k] = PersistentMapping(v)
        logger.info('Done.')

    def _adaptHolidaysWarningMessage(self):
        '''Adapt holidays warning message to be less panicking...'''
        logger.info("Adapting content for message 'Holidays warning'...")
        message = self.portal.get('messages-config').get('holidays_warning')
        text = translate('holidays_warning_message',
                         domain='PloneMeeting',
                         context=self.request)
        message.text = RichTextValue(
            raw=text,
            mimeType='text/html',
            outputMimeType='text/html',
            encoding='utf-8')
        logger.info('Done.')

    def _fixBadHolidayValue(self):
        """A bad value was historically added '2017/2/25', turn it to '2017/12/25'."""
        logger.info("Fix wrong value in ToolPloneMeeting.holidays...")
        storedHolidays = self.tool.getHolidays()
        holidays = []
        for holiday in storedHolidays:
            if holiday['date'] == '2017/2/25':
                holidays.append({'date': '2017/12/25'})
                logger.info('Wrong value "2017/2/25" was fixed!')
            else:
                holidays.append(holiday)
        self.tool.setHolidays(holidays)
        logger.info('Done.')

    def run(self):
        logger.info('Migrating to PloneMeeting 4102...')
        self._updateFTWLabelsStorage()
        self._adaptHolidaysWarningMessage()
        self._fixBadHolidayValue()


def migrate(context):
    '''This migration function will:

       1) Update ftw.labels jar storage to use PersistentMappings;
       2) Update collective.messagesviewlet 'holidays warning' text;
       3) Fix bad value '2017/2/25' in ToolPloneMeeting.holidays.
    '''
    migrator = Migrate_To_4102(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-

from collective.contact.plonegroup.utils import select_organization
from copy import deepcopy
from ftw.labels.interfaces import ILabelJar
from imio.helpers.content import richtextval
from plone import api
from plone.app.testing import logout
from plone.dexterity.utils import createContentInContainer
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.permissions import View
from Products.PloneMeeting.profiles import MeetingConfigDescriptor
from zope.component import getAdapter
from zope.event import notify


class PloneMeetingTestingHelpers(object):
    '''Stub class that provides some helper methods about testing.'''

    TRANSITIONS_FOR_PROPOSING_ITEM_FIRST_LEVEL_1 = TRANSITIONS_FOR_PROPOSING_ITEM_FIRST_LEVEL_2 = ('propose', )
    TRANSITIONS_FOR_PROPOSING_ITEM_1 = TRANSITIONS_FOR_PROPOSING_ITEM_2 = ('propose', 'prevalidate', )

    TRANSITIONS_FOR_VALIDATING_ITEM_1 = TRANSITIONS_FOR_VALIDATING_ITEM_2 = ('propose', 'validate', )
    TRANSITIONS_FOR_PREVALIDATING_ITEM_1 = TRANSITIONS_FOR_PREVALIDATING_ITEM_2 = ('propose', 'prevalidate', )
    TRANSITIONS_FOR_PRESENTING_ITEM_1 = TRANSITIONS_FOR_PRESENTING_ITEM_2 = (
        'propose', 'prevalidate', 'validate', 'present', )

    TRANSITIONS_FOR_PUBLISHING_MEETING_1 = TRANSITIONS_FOR_PUBLISHING_MEETING_2 = ('freeze', 'publish', )
    TRANSITIONS_FOR_FREEZING_MEETING_1 = TRANSITIONS_FOR_FREEZING_MEETING_2 = ('freeze', )
    TRANSITIONS_FOR_DECIDING_MEETING_1 = TRANSITIONS_FOR_DECIDING_MEETING_2 = ('freeze', 'publish', 'decide')
    TRANSITIONS_FOR_CLOSING_MEETING_1 = TRANSITIONS_FOR_CLOSING_MEETING_2 = ('freeze',
                                                                             'publish',
                                                                             'decide',
                                                                             'close', )
    TRANSITIONS_FOR_ACCEPTING_ITEMS_MEETING_1 = ('freeze', 'publish', 'decide')
    TRANSITIONS_FOR_ACCEPTING_ITEMS_MEETING_2 = ('freeze', 'publish', 'decide')
    BACK_TO_WF_PATH_1 = BACK_TO_WF_PATH_2 = {
        # Meeting
        'created': ('backToDecisionsPublished',
                    'backToDecided',
                    'backToPublished',
                    'backToFrozen',
                    'backToCreated',),
        # MeetingItem
        'itemcreated': ('backToItemPublished',
                        'backToItemFrozen',
                        'backToPresented',
                        'backToValidated',
                        'backToPreValidated',
                        'backToProposed',
                        'backToItemCreated', ),
        'proposed': ('backToItemPublished',
                     'backToItemFrozen',
                     'backToPresented',
                     'backToValidated',
                     'backToProposed', ),
        'validated': ('backToItemPublished',
                      'backToItemFrozen',
                      'backToPresented',
                      'backToValidated', ),
        'presented': ('backToItemPublished',
                      'backToItemFrozen',
                      'backToItemPublished',
                      'backToPresented', )}

    WF_ITEM_STATE_NAME_MAPPINGS_1 = {'itemcreated': 'itemcreated',
                                     'proposed_first_level': 'proposed',
                                     'proposed': 'proposed',
                                     'prevalidated': 'prevalidated',
                                     'validated': 'validated',
                                     'presented': 'presented',
                                     'itemfrozen': 'itemfrozen'}
    WF_ITEM_STATE_NAME_MAPPINGS_2 = {'itemcreated': 'itemcreated',
                                     'proposed_first_level': 'proposed',
                                     'proposed': 'proposed',
                                     'prevalidated': 'prevalidated',
                                     'validated': 'validated',
                                     'presented': 'presented',
                                     'itemfrozen': 'itemfrozen'}
    WF_MEETING_STATE_NAME_MAPPINGS_1 = {'frozen': 'frozen'}
    WF_MEETING_STATE_NAME_MAPPINGS_2 = {'frozen': 'frozen'}

    WF_ITEM_TRANSITION_NAME_MAPPINGS_1 = {
        'backToItemCreated': 'backToItemCreated',
        'backToProposed': 'backToProposed', }
    WF_ITEM_TRANSITION_NAME_MAPPINGS_2 = {
        'backToItemCreated': 'backToItemCreated',
        'backToProposed': 'backToProposed', }
    WF_MEETING_TRANSITION_NAME_MAPPINGS_1 = {}
    WF_MEETING_TRANSITION_NAME_MAPPINGS_2 = {}

    # in which state an item must be after a particular meeting transition?
    ITEM_WF_STATE_AFTER_MEETING_TRANSITION = {'publish_decisions': 'accepted',
                                              'close': 'accepted', }

    def _stateMappingFor(self, state_name, meta_type='MeetingItem'):
        """ """
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        pattern = ''
        if meta_type == 'MeetingItem':
            pattern = 'WF_ITEM_STATE_NAME_MAPPINGS_%d'
        else:
            pattern = 'WF_MEETING_STATE_NAME_MAPPINGS_%d'
        return getattr(self, (pattern % meetingConfigNumber)).get(state_name, state_name)

    def _transitionMappingFor(self, transition_name, meta_type='MeetingItem'):
        """ """
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        pattern = ''
        if meta_type == 'MeetingItem':
            pattern = 'WF_ITEM_TRANSITION_NAME_MAPPINGS_%d'
        else:
            pattern = 'WF_MEETING_TRANSITION_NAME_MAPPINGS_%d'
        return getattr(self, (pattern % meetingConfigNumber)).get(transition_name, transition_name)

    def _createMeetingWithItems(self, meetingDate=None):
        '''Create a meeting with a bunch of items.'''
        def _set_proposing_group(item, org):
            """Take into account fact that configuration uses groupsInCharge."""
            groups_in_charge = org.groups_in_charge
            if groups_in_charge:
                item.setProposingGroupWithGroupInCharge(
                    '{0}__groupincharge__{1}'.format(
                        org.UID(), groups_in_charge[0]))
            else:
                item.setProposingGroup(org.UID())
        meeting = self.create('Meeting', date=meetingDate)
        # a meeting could be created with items if it has
        # recurring items...  But we can also add some more...
        # id=item-1
        item1 = self.create('MeetingItem', title='Item 1')
        _set_proposing_group(item1, self.vendors)
        item1.setAssociatedGroups((self.developers_uid,))
        item1.setPrivacy('public')
        item1.setPollType('secret_separated')
        item1.setCategory('research')
        item1.setClassifier('classifier3')
        item1.reindexObject()
        # id=item-2
        item2 = self.create('MeetingItem', title='Item 2')
        _set_proposing_group(item2, self.developers)
        item2.setPrivacy('public')
        item2.setPollType('no_vote')
        item2.setCategory('development')
        item2.setClassifier('classifier2')
        item2.reindexObject()
        # id=item-3
        item3 = self.create('MeetingItem', title='Item 3')
        _set_proposing_group(item3, self.vendors)
        item3.setPrivacy('secret')
        item3.setPollType('freehand')
        item3.setCategory('development')
        item3.setClassifier('classifier2')
        item3.reindexObject()
        # id=item-4
        item4 = self.create('MeetingItem', title='Item 4')
        _set_proposing_group(item4, self.developers)
        item4.setPrivacy('secret')
        item4.setPollType('freehand')
        item4.setCategory('events')
        item4.setClassifier('classifier1')
        item4.reindexObject()
        # id=item-5
        item5 = self.create('MeetingItem', title='Item 5')
        _set_proposing_group(item5, self.vendors)
        item5.setPrivacy('public')
        item5.setPollType('secret')
        item5.setCategory('events')
        item5.setClassifier('classifier1')
        item5.reindexObject()
        for item in (item1, item2, item3, item4, item5):
            item.setDecision('<p>A decision</p>')
            self.presentItem(item)
        return meeting

    def _getTransitionsToCloseAMeeting(self):
        """ """
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        return getattr(self, ('TRANSITIONS_FOR_CLOSING_MEETING_%d' % meetingConfigNumber))

    def _getNecessaryMeetingTransitionsToAcceptItem(self):
        '''Returns the necessary transitions to trigger on the Meeting before being
           able to accept an item.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        return getattr(self, ('TRANSITIONS_FOR_ACCEPTING_ITEMS_MEETING_%d' % meetingConfigNumber))

    def get_transitions_for_proposing_item(self, first_level=False):
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        if first_level:
            return getattr(
                self, ('TRANSITIONS_FOR_PROPOSING_ITEM_FIRST_LEVEL_%d' % meetingConfigNumber))
        else:
            return getattr(
                self, ('TRANSITIONS_FOR_PROPOSING_ITEM_%d' % meetingConfigNumber))

    def proposeItem(self, item, first_level=False, as_manager=True, clean_memoize=True):
        '''Propose passed p_item using TRANSITIONS_FOR_PROPOSING_ITEM_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_PROPOSING_ITEM_x is 1 or 2.
           If p_first_level is True, we will use TRANSITIONS_FOR_PROPOSING_ITEM_FIRST_LEVEL_x,
           this makes it possible to reach an intermediate propose level.'''
        self._doTransitionsFor(item,
                               self.get_transitions_for_proposing_item(first_level),
                               as_manager=as_manager,
                               clean_memoize=clean_memoize)

    def prevalidateItem(self, item, as_manager=True, clean_memoize=True):
        '''Prevalidate passed p_item using TRANSITIONS_FOR_PREVALIDATING_ITEM_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_PREVALIDATING_ITEM_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            item,
            getattr(self, ('TRANSITIONS_FOR_PREVALIDATING_ITEM_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def validateItem(self, item, as_manager=True, clean_memoize=True):
        '''Validate passed p_item using TRANSITIONS_FOR_VALIDATING_ITEM_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_VALIDATING_ITEM_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            item,
            getattr(self, ('TRANSITIONS_FOR_VALIDATING_ITEM_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def presentItem(self, item, as_manager=True, clean_memoize=True):
        '''Present passed p_item using TRANSITIONS_FOR_PRESENTING_ITEM_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_PRESENTING_ITEM_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            item,
            getattr(self, ('TRANSITIONS_FOR_PRESENTING_ITEM_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def publishMeeting(self, meeting, as_manager=False, clean_memoize=True):
        '''Publish passed p_meeting using TRANSITIONS_FOR_PUBLISHING_MEETING_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_PUBLISHING_MEETING_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            meeting,
            getattr(self, ('TRANSITIONS_FOR_PUBLISHING_MEETING_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def freezeMeeting(self, meeting, as_manager=False, clean_memoize=True):
        '''Freeze passed p_meeting using TRANSITIONS_FOR_FREEZING_MEETING_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_FREEZING_MEETING_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            meeting,
            getattr(self, ('TRANSITIONS_FOR_FREEZING_MEETING_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def decideMeeting(self, meeting, as_manager=False, clean_memoize=True):
        '''Decide passed p_meeting using TRANSITIONS_FOR_DECIDING_MEETING_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_DECIDING_MEETING_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            meeting,
            getattr(self, ('TRANSITIONS_FOR_DECIDING_MEETING_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def apply_meeting_transition_to_late_state(self, meeting, as_manager=False, clean_memoize=True):
        self.freezeMeeting(meeting, as_manager, clean_memoize)

    def closeMeeting(self, meeting, as_manager=False, clean_memoize=True):
        '''Close passed p_meeting using TRANSITIONS_FOR_CLOSING_MEETING_x.
           The p_meetingConfigNumber specify if we use meetingConfig or meetingConfig2, so
           the _x here above in TRANSITIONS_FOR_CLOSING_MEETING_x is 1 or 2.'''
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        self._doTransitionsFor(
            meeting,
            getattr(self, ('TRANSITIONS_FOR_CLOSING_MEETING_%d' % meetingConfigNumber)),
            as_manager=as_manager,
            clean_memoize=clean_memoize)

    def backToState(self, itemOrMeeting, state, as_manager=True, comment=""):
        """Set the p_item back to p_state.
           Given p_state MUST BE original state name, aka state existing in PloneMeeting workflow."""
        # if a wf path is defined in BACK_TO_WF_PATH_x to go to relevant state, use it
        # if not, trigger every 'backToXXX' existing transition
        meetingConfigNumber = self._determinateUsedMeetingConfigNumber()
        BACK_TO_WF_PATH = getattr(self, 'BACK_TO_WF_PATH_%d' % meetingConfigNumber)
        useDefinedWfPath = False
        # check if a mapping exist for state name, returns original state if no mapping exists
        state = self._stateMappingFor(state)
        if state in BACK_TO_WF_PATH:
            transitions = BACK_TO_WF_PATH[state]
            useDefinedWfPath = True
        # do things as admin to avoid permission issues
        if as_manager:
            currentUser = self.member.getId()
            self.changeUser('admin')
        max_attempts = 20
        nb_attempts = 0
        while not itemOrMeeting.query_state() == state and nb_attempts <= max_attempts:
            nb_attempts += 1
            if not useDefinedWfPath:
                transitions = self.transitions(itemOrMeeting)
            for tr in transitions:
                if (tr.startswith('back') or useDefinedWfPath) and tr in self.transitions(itemOrMeeting):
                    self.do(itemOrMeeting, tr, comment=comment)
                    break
        if nb_attempts >= max_attempts or itemOrMeeting.query_state() != state:
            raise ValueError('impossible to go back to {}'.format(state))
        if as_manager:
            self.changeUser(currentUser)

    def _doTransitionsFor(self, itemOrMeeting, transitions, as_manager=False, comment="", clean_memoize=True):
        """Helper that just trigger given p_transitions on given p_itemOrMeeting."""
        # do things as admin to avoid permission issues
        self.request['disable_check_required_data'] = True
        if as_manager:
            currentUser = self.member.getId()
            self.changeUser('admin', clean_memoize=clean_memoize)
        for tr in transitions:
            if tr in self.transitions(itemOrMeeting):
                self.do(itemOrMeeting, tr, comment=comment, clean_memoize=clean_memoize)
        if as_manager:
            self.changeUser(currentUser, clean_memoize=clean_memoize)
        self.request['disable_check_required_data'] = False

    def _determinateUsedMeetingConfigNumber(self):
        """Helper method that check if we use meetingConfig or meetingConfig2."""
        if self.meetingConfig.getId() == self.meetingConfig2.getId():
            # we are using meetingConfig2
            return 2
        return 1

    def _removeAllMembers(self, group, members):
        """Allow to remove all members from a group.
           Overrided to do it as 'Manager' to be able not
           to change permissions ever lines"""
        from plone.app.testing.helpers import setRoles
        currentMember = api.user.get_current()
        currentMemberRoles = tuple(currentMember.getRoles())
        setRoles(self.portal, currentMember.getId(), currentMemberRoles + ('Manager', ))
        for member in members:
            group.removeMember(member)
        setRoles(self.portal, currentMember.getId(), currentMemberRoles)

    def _remove_all_members_from_groups(self, group_id_list):
        for group in group_id_list:
            self._removeAllMembers(
                api.group.get(group),
                api.group.get(group).getMemberIds())

    def _addAllMembers(self, group, members):
        """Allow to add again all the members from a group.
           Overrided to do it as 'Manager' to be able not
           to change permissions ever lines"""
        from plone.app.testing.helpers import setRoles
        currentMember = api.user.get_current()
        currentMemberRoles = currentMember.getRoles()
        setRoles(self.portal, currentMember.getId(), currentMemberRoles + ['Manager', ])
        for member in members:
            group.addMember(member)
        setRoles(self.portal, currentMember.getId(), currentMemberRoles)

    def _removeUsersFromEveryGroups(self, user_ids):
        """Remove extra users from their groups to not break test,
           useful for profiles having extra users or roles."""
        for extra_user_id in user_ids:
            user = api.user.get(extra_user_id)
            # remove from every groups, bypass Plone groups (including virtual)
            for group_id in [user_group_id for user_group_id in user.getGroups()
                             if '_' in user_group_id]:
                api.group.remove_user(groupname=group_id, username=extra_user_id)

    def _initial_state(self, obj):
        """Return the workflow initial_state of given p_obj."""
        wf_name = self.wfTool.getWorkflowsFor(obj)[0].getId()
        return self.wfTool[wf_name].initial_state

    def _make_not_found_user(self, user_id='new_test_user'):
        """Add a p_user_id member to the p_group_id group then delete it."""
        currentUser = self.member.getId()
        self.changeUser('admin')
        membershipTool = api.portal.get_tool('portal_membership')
        membershipTool.addMember(id=user_id,
                                 password='12345',
                                 roles=('Member', ),
                                 domains=())
        self._addPrincipalToGroup(user_id, self.developers_creators)
        membershipTool.deleteMembers((user_id, ))
        # now we have a 'not found' user in developers_creators
        self.assertTrue((user_id, '<{0}: not found>'.format(user_id)) in
                        self.portal.acl_users.source_groups.listAssignedPrincipals(
                            self.developers_creators))
        # groupData.getGroupMembers/groupData.getGroupMemberIds ignore not found
        self.assertFalse(user_id in api.group.get(self.developers_creators).getGroupMemberIds())
        self.changeUser(currentUser)

    def _check_access(self, obj, userIds=[], read=True, write=True):
        """ """
        original_user_id = self.member.getId()
        # no userIds means use current user id
        if not userIds:
            userIds = [original_user_id]
        for userId in userIds:
            self.changeUser(userId)
            if read:
                self.assertTrue(self.hasPermission(View, obj),
                                'Should have "View" on obj')
            else:
                self.assertFalse(self.hasPermission(View, obj),
                                 'Should not "View" on obj')
            if write:
                self.assertTrue(self.hasPermission(ModifyPortalContent, obj),
                                'Should have "Modify" on obj')
            else:
                self.assertFalse(self.hasPermission(ModifyPortalContent, obj),
                                 'Should not "Modify" on obj')
        self.changeUser(original_user_id)

    def _setupStorePodAsAnnex(self):
        """ """
        cfg = self.meetingConfig
        pod_template = cfg.podtemplates.itemTemplate
        annex_type = cfg.annexes_types.item_annexes.get('item-annex')
        pod_template.store_as_annex = annex_type.UID()
        self.request.set('template_uid', pod_template.UID())
        self.request.set('output_format', 'odt')
        self.request.set('store_as_annex', '1')
        # create an item
        original_member_id = self.member.getId()
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        if original_member_id:
            self.changeUser(original_member_id)
        return pod_template, annex_type, item

    def _setupItemWithAdvice(self):
        """Create an item and adds an advice to it."""
        cfg = self.meetingConfig
        cfgItemWF = self.wfTool.getWorkflowsFor(cfg.getItemTypeName())[0]
        item_initial_state = self.wfTool[cfgItemWF.getId()].initial_state

        cfg.setItemAdviceStates((item_initial_state, ))
        cfg.setItemAdviceEditStates((item_initial_state, ))
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        item.setOptionalAdvisers((self.vendors_uid, ))
        item.update_local_roles()
        self.changeUser('pmReviewer2')
        advice = createContentInContainer(
            item,
            item.adapted()._advicePortalTypeForAdviser(self.vendors_uid),
            **{'advice_group': self.vendors_uid,
               'advice_type': u'positive',
               'advice_comment': richtextval(u'My comment')})
        return item, advice

    def _setUpGroupsInCharge(self, item, groups=[]):
        """As group in charge is an adaptable method, it may be setup differently."""
        if not groups:
            groups = [self.vendors_uid]
        item.setGroupsInCharge(groups)
        item.update_local_roles()

    def _setUpCommitteeEditor(self, cfg):
        """Will setup "pmCreator2" as committee_1 editor."""
        cfg_committees = cfg.getCommittees()
        cfg_committees[0]['enable_editors'] = "1"
        cfg.setCommittees(cfg_committees)
        notify(ObjectEditedEvent(cfg))
        self._addPrincipalToGroup(
            'pmCreator2', "{0}_{1}".format(cfg.getId(), 'committee_1'))

    def _tearDownGroupsInCharge(self, item):
        """If group in charge is overrided, it may be setup differently."""
        item.setGroupsInCharge([])

    def _select_organization(self, org_uid, remove=False):
        """Select organization in ORGANIZATIONS_REGISTRY."""
        select_organization(org_uid, remove=remove)
        self.cleanMemoize()

    def _enablePrevalidation(self, cfg, enable_extra_suffixes=False):
        """Enable the 'prevalidated' state in MeetingConfig.itemWFValidationLevels."""
        currentUser = self.member.getId()
        self.changeUser('admin')
        itemWFValidationLevels = cfg.getItemWFValidationLevels()
        itemWFValidationLevels[1]['suffix'] = 'prereviewers'
        itemWFValidationLevels[2]['enabled'] = '1'
        if enable_extra_suffixes:
            itemWFValidationLevels[2]['extra_suffixes'] = ['reviewers']
        cfg.setItemWFValidationLevels(itemWFValidationLevels)
        notify(ObjectEditedEvent(cfg))
        self.changeUser(currentUser)

    def _updateItemValidationLevel(self, cfg, level=None, enable=True, **kwargs):
        """Utility method that enable/disable item validation levels."""
        currentUser = self.member.getId()
        self.changeUser('admin')
        itemValLevels = cfg.getItemWFValidationLevels()
        for itemValLevel in itemValLevels:
            if not level or itemValLevel['state'] == level:
                itemValLevel['enabled'] = enable and '1' or '0'
                for k in itemValLevel.keys():
                    if k in kwargs:
                        itemValLevel[k] = kwargs[k]
        cfg.setItemWFValidationLevels(itemValLevels)
        notify(ObjectEditedEvent(cfg))
        self.changeUser(currentUser)

    def _enableItemValidationLevel(self, cfg, level=None, suffix=None):
        """Enable one or every item validation levels."""
        kwargs = {}
        if level is not None:
            kwargs['level'] = level
        if suffix is not None:
            kwargs['suffix'] = suffix
        self._updateItemValidationLevel(cfg, enable=True, **kwargs)

    def _disableItemValidationLevel(self, cfg, level=None, suffix=None):
        """Disable one or every item validation levels."""
        kwargs = {}
        if level is not None:
            kwargs['level'] = level
        if suffix is not None:
            kwargs['suffix'] = suffix
        self._updateItemValidationLevel(cfg, enable=False, **kwargs)

    def _setUpOrderedContacts(
            self,
            meeting_attrs=('attendees', 'excused', 'absents', 'signatories', ),
            item_attrs=('itemInitiator', )):
        """ """
        # login to be able to query held_positions for orderedContacts vocabulary
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        cfg.setUsedMeetingAttributes(meeting_attrs)
        cfg.setUsedItemAttributes(item_attrs)
        ordered_contacts = cfg.getField('orderedContacts').Vocabulary(cfg).keys()
        cfg.setOrderedContacts(ordered_contacts)
        logout()

    def _setItemToWaitingAdvices(self, item, transition):
        """Done to be overrided, sometimes it is necessary to do something more to be able
           to set item to 'waiting_advices'."""
        self.do(item, transition)

    def _userAbleToBackFromWaitingAdvices(self, currentState):
        """Return username able to back from waiting advices."""
        if currentState == 'itemcreated_waiting_advices':
            return 'pmCreator1'
        else:
            return 'pmReviewer1'

    def _afterItemCreatedWaitingAdviceWithPrevalidation(self, item):
        """Made to be overrided..."""
        return

    def _setUpCommittees(self, attendees=True):
        """Setup use of committees with attendees/signatories or assembly/signatures."""
        # avoid circular import problems
        from Products.PloneMeeting.content.meeting import default_committees
        from Products.PloneMeeting.tests.PloneMeetingTestCase import DefaultData
        cfg = self.meetingConfig
        if attendees:
            to_enable = ["committees", "committees_attendees", "committees_signatories"]
            to_disable = ["committees_assembly", "committees_signatures"]
        else:
            to_enable = ["committees", "committees_assembly", "committees_signatures"]
            to_disable = ["committees_attendees", "committees_signatories"]
        self._enableField(to_enable, related_to="Meeting")
        self._enableField(to_disable, related_to="Meeting", enable=False)
        cfg.setOrderedCommitteeContacts((self.hp1_uid, self.hp2_uid, self.hp3_uid))
        cfg_committees = cfg.getCommittees()
        cfg_committees[0]['default_assembly'] = "Default assembly"
        cfg_committees[0]['default_signatures'] = "Line 1,\r\nLine 2\r\nLine 3,\r\nLine 4"
        cfg_committees[0]['default_place'] = "Default place"
        cfg_committees[0]['default_attendees'] = [self.hp1_uid, self.hp2_uid]
        cfg_committees[0]['default_signatories'] = [self.hp2_uid, self.hp3_uid]
        cfg.setCommittees(cfg_committees)
        meeting = self.create('Meeting', committees=default_committees(DefaultData(cfg)))
        meeting.committees[0]['committee_observations'] = richtextval('<p>Committee observations</p>')
        return meeting

    def _setUpDefaultItemWFValidationLevels(self, cfg):
        """Setup default itemWFValidationLevels for given p_cfg,
           used to avoid a custom profile breaking the tests."""
        # make sure we use default itemWFValidationLevels,
        # useful when test executed with custom profile
        defValues = MeetingConfigDescriptor.get()
        cfg.setItemWFValidationLevels(deepcopy(defValues.itemWFValidationLevels))
        notify(ObjectEditedEvent(cfg))

    def _setupLabelsEditableWhenItemEditable(self, cfg, enable=True):
        """Setup labels only editable when item editable."""
        self._enableField('labels')
        labelsConfig = cfg.getLabelsConfig()
        if enable:
            labelsConfig[0]['edit_groups'] = []
            labelsConfig[0]['edit_access_on'] = 'python: cfg.isManager(cfg) or '\
                'checkPermission("Modify portal content", context)'
            labelsConfig[0]['edit_access_on_cache'] = '0'
        else:
            labelsConfig[0]['edit_groups'] = [
                'suffix_proposing_group_creators',
                'suffix_proposing_group_prereviewers',
                'suffix_proposing_group_reviewers']
            labelsConfig[0]['edit_access_on'] = ""
            labelsConfig[0]['edit_access_on_cache'] = '1'
        cfg.setLabelsConfig(labelsConfig)

    def _enable_ftw_labels(self, cfg, add_follow_up=False):
        self._enableField('labels')
        self.changeUser('pmCreator1')
        labeljar = getAdapter(cfg, ILabelJar)
        labeljar.add('Label1', 'green', False)
        labeljar.add('Label2', 'red', False)
        if add_follow_up:
            labeljar.add('Needed follow-up', 'orange', False)
            labeljar.add('Provided follow-up', 'green-light', False)
        return labeljar

    def _setupFollowUp(self, cfg):
        """Configure followUp labels."""
        self._enable_ftw_labels(cfg, add_follow_up=True)
        self._enableField(['neededFollowUp', 'providedFollowUp'])
        config = list(cfg.getLabelsConfig())
        # needed-follow-up
        new_config = deepcopy(config[0])
        new_config['label_id'] = "needed-follow-up"
        new_config['edit_groups'] = ["configgroup_meetingmanagers"]
        config.append(new_config)
        # provided-follow-up
        new_config = deepcopy(config[0])
        new_config['label_id'] = "provided-follow-up"
        new_config['edit_access_on'] = "python: not utils.fieldIsEmpty('providedFollowUp', item)"
        new_config['edit_access_on_cache'] = "0"
        config.append(new_config)
        cfg.setLabelsConfig(config)



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from plone import api
from Products.Five.browser import BrowserView
from Products.PloneMeeting.config import PloneMeetingError
from Products.PloneMeeting.utils import ItemListTypeChangedEvent
from Products.PloneMeeting.utils import notifyModifiedAndReindex
from zope.component import queryUtility
from zope.event import notify
from zope.schema.interfaces import IVocabularyFactory


class ItemListTypeView(BrowserView):
    '''Render the item listType selection on the meetingitem_view.'''

    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.context = context
        self.request = request
        self.portal = api.portal.get_tool('portal_url').getPortalObject()
        self.portal_url = self.portal.absolute_url()
        factory = queryUtility(IVocabularyFactory,
                               'Products.PloneMeeting.vocabularies.listtypesvocabulary')
        self.vocab = factory(self.context)

    def selectableListTypes(self):
        '''Returns a list of listTypes the current user can set the item to.'''
        if not self.context.adapted().mayChangeListType():
            return []
        list_type = self.context.getListType()
        return [term for term in self.vocab if term.value != list_type]

    def force_faceted(self):
        """ """
        on_meeting = self.context.wfConditions()._publishedObjectIsMeeting()
        return on_meeting and 'true' or 'false'

    def currentListTypeTitle(self):
        """ """
        return self.vocab.getTerm(self.context.getListType()).title

    def displayCurrentListType(self):
        '''Display current listType?  We display it on the item view but not on the meeting_view
           in the itemListType column.'''
        return bool(self.request.getURL().startswith(self.context.absolute_url()))


class ChangeItemListTypeView(BrowserView):
    '''This manage the item listType changes.'''

    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.context = context
        self.request = request

    def __call__(self, new_value):
        '''Change listType value.'''
        self._changeListType(new_value)

    def _changeListType(self, new_value):
        '''Helper method that changes listType value and check that :
           - new_value is among selectable listType values;
           - user actually mayChangeListType.'''
        # make sure new_value exists
        factory = queryUtility(IVocabularyFactory,
                               'Products.PloneMeeting.vocabularies.listtypesvocabulary')
        if new_value not in factory(self.context):
            raise KeyError("New value '{0}' does not correspond to a value of MeetingItem.listType".
                           format(new_value))

        if not self.context.adapted().mayChangeListType():
            raise Unauthorized

        # save old_listType so we can pass it the the ItemListTypeChangedEvent
        # set the new listType and notify events
        old_listType = self.context.getListType()
        self.context.setListType(new_value)
        self.context._update_after_edit(idxs=['listType'])
        if new_value == 'late':
            self.context.send_powerobservers_mail_if_relevant('late_item_in_meeting')

        try:
            notify(ItemListTypeChangedEvent(self.context, old_listType))
        except PloneMeetingError, msg:
            # back to original state
            self.context.setListType(old_listType)
            self.context._update_after_edit(idxs=['listType'])
            plone_utils = api.portal.get_tool('plone_utils')
            plone_utils.addPortalMessage(msg, type='warning')

        # an item's listType has been changed, notify meeting
        if self.context.hasMeeting():
            meeting = self.context.getMeeting()
            notifyModifiedAndReindex(meeting)



# -*- coding: utf-8 -*-

from DateTime import DateTime
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.PloneMeeting.utils import get_public_url


class Migrate_To_4203(Migrator):

    def _adaptInternalImagesLinkToUseResolveUID(self):
        """We make sure we use resolveuid in src to internal images."""
        logger.info('Adapting link to internal images to use resolveuid...')
        # as bug was introduced in version 4.2, update items that were updated after
        # in portal_setup logs, collective.fontawesome first appearance is when applying
        # upgrade step to 4200, we use it as base modified time query
        ps_logs = sorted([log for log in self.ps.objectValues()
                          if 'fontawesome' in log.__name__])
        install_time = DateTime('1990/01/01')
        if ps_logs:
            install_time = ps_logs[0].bobobase_modification_time()
        # take only items modified since upgrade to 4200
        brains = self.catalog(meta_type='MeetingItem',
                              modified={'range': 'min', 'query': install_time})
        i = 1
        total = len(brains)
        number_of_migrated_links = 0
        for brain in brains:
            logger.info('Migrating links to image {0}/{1} ({2})...'.format(
                i,
                total,
                brain.getPath()))
            i = i + 1
            item = brain.getObject()
            for image in [obj for obj in item.objectValues() if obj.portal_type == 'Image']:
                # get image url taking env var PUBLIC_URL into account
                image_url = get_public_url(image)
                image_UID = image.UID()
                for field in item.Schema().filterFields(default_content_type='text/html'):
                    content = field.getRaw(item)
                    if content.find(image_url) != -1:
                        content = content.replace(image_url, 'resolveuid/{0}'.format(image_UID))
                        logger.info('Replaced image link in field {0}'.format(field.getName()))
                        number_of_migrated_links = number_of_migrated_links + 1
                        field.set(item, content)
        logger.info('Adapted {0} links.'.format(number_of_migrated_links))
        logger.info('Done.')

    def run(self, extra_omitted=[], from_migration_to_4200=False):

        logger.info('Migrating to PloneMeeting 4203...')

        # not necessary if executing the full upgrade to 4200
        # as problem was introduced after 4200...
        if not from_migration_to_4200:
            self._adaptInternalImagesLinkToUseResolveUID()
        logger.info('Done.')


def migrate(context):
    '''This migration function will:

       1) Adapt link to images to ensure it uses resolveuid.
    '''
    migrator = Migrate_To_4203(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-

from collective.contact.plonegroup.utils import get_own_organization
from collective.contact.widget.source import ContactSource
from collective.contact.widget.source import ContactSourceBinder


#############################################
#
#      NOT USED ANYMORE TO BE REMOVED
#
#############################################

class PMContactSource(ContactSource):
    """Returns organizations, except ones stored in PLONEGROUP_ORG."""

    def search(self, query, relations=None, limit=50):
        """Base selectable_filter do the job but do not query own_org,
           we append it at the beginning of the vocabulary."""
        results = [term for term in super(PMContactSource, self).search(query, relations, limit)]
        brains = self.catalog(UID=get_own_organization().UID())
        results.insert(0, self.getTermByBrain(brains[0], real_value=False))
        return results


class PMContactSourceBinder(ContactSourceBinder):
    """Returns organizations, except ones stored in PLONEGROUP_ORG."""
    path_source = PMContactSource

    def __init__(self, navigation_tree_query=None, default=None, defaultFactory=None, **kw):
        super(PMContactSourceBinder, self).__init__(navigation_tree_query, default, defaultFactory, **kw)
        self.selectable_filter.criteria = {
            'portal_type': ('organization', ),
            'object_provides': ['collective.contact.plonegroup.interfaces.INotPloneGroupContact']}



# -*- coding: utf-8 -*-

from collective.iconifiedcategory.content.category import CategorySchemaPolicy
from collective.iconifiedcategory.content.subcategory import SubcategorySchemaPolicy
from imio.helpers.content import uuidToObject
from natsort import humansorted
from operator import attrgetter
from plone import api
from plone.autoform import directives as form
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.widgets.pm_checkbox import PMCheckBoxFieldWidget
from z3c.form.browser.radio import RadioFieldWidget
from zope import schema
from zope.globalrequest import getRequest
from zope.i18n import translate
from zope.interface import implements
from zope.interface import Interface
from zope.interface import Invalid
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


def other_mc_correspondences_constraint(values):
    ''' '''
    if values:
        tool = api.portal.get_tool('portal_plonemeeting')
        previous_cfg_id = None
        for value in values:
            cfg = tool.getMeetingConfig(uuidToObject(value, unrestricted=True))
            if cfg is None:
                # case of meeting-config-id.annex_not_kept
                cfg_id = value.split('.')[0]
            else:
                cfg_id = cfg.getId()
            if cfg_id == previous_cfg_id:
                msg = translate(u'error_can_not_select_several_values_for_same_mc',
                                domain="PloneMeeting",
                                context=getRequest())
                raise Invalid(msg)
            previous_cfg_id = cfg_id
    return True


class IPMContentCategory(Interface):
    """
    """

    after_scan_change_annex_type_to = schema.Choice(
        title=_(u'after_scan_change_annex_type_to_title'),
        description=_(u"after_scan_change_annex_type_to_descr"),
        vocabulary="Products.PloneMeeting.vocabularies.item_annex_types_vocabulary",
        required=False,
    )


class IItemAnnexContentCategory(IPMContentCategory):
    """
    """

    form.widget('other_mc_correspondences',
                PMCheckBoxFieldWidget,
                multiple='multiple')
    other_mc_correspondences = schema.Set(
        title=_("PloneMeeting_label_otherMCCorrespondences"),
        description=_("other_mc_correspondences_descr"),
        value_type=schema.Choice(
            vocabulary="Products.PloneMeeting.content.item_annex_content_category."
            "other_mc_correspondences_vocabulary"),
        required=False,
        constraint=other_mc_correspondences_constraint,
    )

    form.widget('only_for_meeting_managers', RadioFieldWidget)
    only_for_meeting_managers = schema.Bool(
        title=_(u'only_for_meeting_managers'),
        description=_(u'only_for_meeting_managers_descr'),
        default=False,
        required=False,
    )


ANNEX_NOT_KEPT = "{0}.annex_not_kept"


class OtherMCCorrespondencesVocabulary(object):
    implements(IVocabularyFactory)

    def __call__(self, context):
        tool = api.portal.get_tool('portal_plonemeeting')
        current_cfg = tool.getMeetingConfig(context)
        res = []
        if current_cfg:
            current_cfg_id = current_cfg.getId()
            for cfg in tool.objectValues('MeetingConfig'):
                cfg_id = cfg.getId()
                if current_cfg_id == cfg_id:
                    continue
                # add a special value to not keep the annex
                res.append(SimpleTerm(
                    ANNEX_NOT_KEPT.format(cfg_id),
                    ANNEX_NOT_KEPT.format(cfg_id),
                    u'%s ➔ %s ➔ %s' % (
                        safe_unicode(cfg.Title(include_config_group=True)),
                        translate('Item annexes',
                                  domain='PloneMeeting',
                                  context=context.REQUEST),
                        translate('do_not_keep_annex',
                                  domain='PloneMeeting',
                                  context=context.REQUEST))))
                for annexes_types_folder in (
                        cfg.annexes_types.item_annexes,
                        cfg.annexes_types.item_decision_annexes):
                    for cat in annexes_types_folder.objectValues():
                        cat_uid = cat.UID()
                        res.append(SimpleTerm(
                            cat_uid,
                            cat_uid,
                            u'%s ➔ %s ➔ %s' % (
                                safe_unicode(cfg.Title(include_config_group=True)),
                                safe_unicode(annexes_types_folder.Title()),
                                safe_unicode(cat.Title()))))
                        for subcat in cat.objectValues():
                            subcat_uid = subcat.UID()
                            res.append(SimpleTerm(
                                subcat_uid,
                                subcat_uid,
                                u'%s ➔ %s ➔ %s ➔ %s' % (
                                    safe_unicode(cfg.Title(include_config_group=True)),
                                    safe_unicode(annexes_types_folder.Title()),
                                    safe_unicode(cat.Title()),
                                    safe_unicode(subcat.Title()))))
        return SimpleVocabulary(humansorted(res, key=attrgetter('title')))


OtherMCCorrespondencesVocabularyFactory = OtherMCCorrespondencesVocabulary()


class ItemAnnexContentCategorySchemaPolicy(CategorySchemaPolicy):

    def bases(self, schema_name, tree):
        bases = super(ItemAnnexContentCategorySchemaPolicy, self).bases(schema_name, tree)
        return bases + (IItemAnnexContentCategory, )


class ItemAnnexContentSubcategorySchemaPolicy(SubcategorySchemaPolicy):

    def bases(self, schema_name, tree):
        bases = super(ItemAnnexContentSubcategorySchemaPolicy, self).bases(schema_name, tree)
        return bases + (IItemAnnexContentCategory, )


class PMContentCategorySchemaPolicy(CategorySchemaPolicy):

    def bases(self, schema_name, tree):
        bases = super(PMContentCategorySchemaPolicy, self).bases(schema_name, tree)
        return bases + (IPMContentCategory, )


class PMContentSubcategorySchemaPolicy(SubcategorySchemaPolicy):

    def bases(self, schema_name, tree):
        bases = super(PMContentSubcategorySchemaPolicy, self).bases(schema_name, tree)
        return bases + (IPMContentCategory, )



# -*- coding: utf-8 -*-
# GNU General Public License (GPL)
'''This module allows to perform some standard sets of adaptations in the
   PloneMeeting data structures and workflows.'''

from imio.helpers.content import get_vocab_values
from imio.helpers.content import object_values
from imio.pyutils.utils import merge_dicts
from imio.pyutils.utils import replace_in_list
from plone import api
from Products.CMFCore.permissions import AccessContentsInformation
from Products.CMFCore.permissions import DeleteObjects
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.permissions import View
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting import logger
from Products.PloneMeeting.config import AddAnnex
from Products.PloneMeeting.config import MEETING_REMOVE_MOG_WFA
from Products.PloneMeeting.utils import updateCollectionCriterion
from zope.i18n import translate


# states of the meeting from which an item can be 'returned_to_proposing_group'
RETURN_TO_PROPOSING_GROUP_FROM_ITEM_STATES = (
    'presented', 'itemfrozen', 'itempublished', 'itemdecided')
# mapping definintions regarding the 'return_to_proposing_group' wfAdaptation
# this is used in MeetingItem.mayBackToMeeting and may vary upon used workflow
# the key is the transition triggerable on the item and the values are states
# of the linked meeting in wich this transition can be triggered
# the last key 'no_more_returnable_state' specify states in wich the item is no more
# returnable to the meeting...
# these mappings are easily overridable by a subproduct...
RETURN_TO_PROPOSING_GROUP_MAPPINGS = {
    'backTo_presented_from_returned_to_proposing_group':
    {'*': ['created']},
    'backTo_itempublished_from_returned_to_proposing_group':
    {'*': ['published']},
    'backTo_itemfrozen_from_returned_to_proposing_group':
    {'*': ['frozen', 'decided', 'decisions_published']},
    'backTo_itemdecided_from_returned_to_proposing_group':
    {'*': ['decided', 'decisions_published']},
    'NO_MORE_RETURNABLE_STATES': ['closed', 'archived']
}

WF_APPLIED = 'Workflow adaptation "%s" applied for meetingConfig "%s".'
WF_APPLIED_CUSTOM = 'Custom Workflow adaptation "%s" applied for meetingConfig "%s".'
ADVICE_WF_APPLIED_CUSTOM = 'Custom advice Workflow adaptation "%s" applied on WF "%s" for every meetingConfigs.'
WF_DOES_NOT_EXIST_WARNING = "Could not apply workflow adaptations because the workflow '%s' does not exist."
WF_ITEM_VALIDATION_LEVELS_DISABLED = 'No enabled item validation levels found for meetingConfig "%s".'

# list of dict containing infos about 'waiting_advices' state(s) to add
# a state will be added by "dict", 'from_states' are list of states leading to the new state
# 'back_states' are states to come back from the new state and 'perm_cloned_state' is the state
# to use to define permissions of the new state minus every 'edit' permissions
WAITING_ADVICES_FROM_STATES = {
    '*':
    (
        {'from_states': ('itemcreated', ),
         'back_states': ('itemcreated', ),
         'perm_cloned_state': 'validated',
         'use_custom_icon': False,
         # default to "validated", this avoid using the backToValidated title that
         # is translated to "Remove from meeting"
         'use_custom_back_transition_title_for': ("validated", ),
         # we can define some back transition id for some back_to_state
         # if not, a generated transition is used, here we could have for example
         # 'defined_back_transition_ids': {"validated": "validate"}
         'defined_back_transition_ids': {},
         # if () given, a custom transition icon is used for every back transitions
         'only_use_custom_back_transition_icon_for': ("validated", ),
         'use_custom_state_title': True,
         'use_custom_transition_title_for': {},
         'remove_modify_access': True,
         'adviser_may_validate': False,
         # must end with _waiting_advices
         'new_state_id': None,
         },
        {'from_states': ('proposed', 'prevalidated'),
         'back_states': ('proposed', 'prevalidated'),
         'perm_cloned_state': 'validated',
         'use_custom_icon': False,
         # is translated to "Remove from meeting"
         'use_custom_back_transition_title_for': ("validated", ),
         # we can define some back transition id for some back_to_state
         # if not, a generated transition is used, here we could have for example
         # 'defined_back_transition_ids': {"validated": "validate"}
         'defined_back_transition_ids': {},
         # if () given, a custom transition icon is used for every back transitions
         'only_use_custom_back_transition_icon_for': ("validated", ),
         'use_custom_state_title': True,
         'use_custom_transition_title_for': {},
         'remove_modify_access': True,
         'adviser_may_validate': False,
         # must end with _waiting_advices
         'new_state_id': None,
         },
    ),
}
# defined here to be importable
WAITING_ADVICES_FROM_TRANSITION_ID_PATTERN = 'wait_advices_from_{0}'
WAITING_ADVICES_FROM_TO_TRANSITION_ID_PATTERN = 'wait_advices_from_{0}__to__{1}'

# restrict item validation back shortcuts
# if not empty, we will permit back shortcuts from given item states
RESTRICT_ITEM_BACK_SHORTCUTS = {
    '*':
    {
        # this means we enable every shortcuts if not defined in other values
        # removing '*': '*' would mean nothing by default
        '*': '*',
    },
}


def get_allowed_back_shortcut_from(cfg_id, state_id):
    """ """
    infos = RESTRICT_ITEM_BACK_SHORTCUTS.get(
        cfg_id, RESTRICT_ITEM_BACK_SHORTCUTS.get('*'))
    allowed_from = infos.get(state_id, infos.get('*'))
    return allowed_from


def get_waiting_advices_infos(cfg_id):
    """ """
    return WAITING_ADVICES_FROM_STATES.get(
        cfg_id, WAITING_ADVICES_FROM_STATES.get('*', ()))


def removePermission(state, perm, role):
    '''For a given p_state, this function ensures that p_role is not more
       among roles who are granted p_perm. '''
    roles = state.permission_roles[perm]
    if role in roles:
        roles = list(roles)
        roles.remove(role)
        state.setPermission(perm, 0, roles)


def grantPermission(state, perm, role):
    '''For a given p_state, this function ensures that p_role is among roles
       who are granted p_perm.'''
    roles = state.permission_roles[perm]
    if role not in roles:
        roles = list(roles)
        roles.append(role)
        state.setPermission(perm, 0, roles)


def change_transition_new_state_id(wf_id, transition_id, new_state_id):
    '''Change given p_new_state_id for p_transition_id.'''
    wfTool = api.portal.get_tool('portal_workflow')
    wf = wfTool.getWorkflowById(wf_id)
    # transition does not exist?
    if transition_id not in wf.transitions:
        logger.error('Transition "{0}" does not exist in WF "{1}"!'.format(transition_id, wf_id))
        return
    # new_state_id does not exist?
    if new_state_id not in wf.states:
        logger.error('New state "{0}" does not exist in WF "{1}"!'.format(new_state_id, wf_id))
        return

    wf.transitions[transition_id].new_state_id = new_state_id
    logger.info('Transition "{0}" new_state_id is now "{1}" in WF "{2}"'.format(
        transition_id, new_state_id, wf_id))


def addState(wf_id,
             new_state_id,
             new_state_title,
             permissions_cloned_state_id,
             leading_transition_id=None,
             leading_transition_title=None,
             back_transitions=[],
             leaving_transition_id=None,
             leaving_to_state_id=None,
             existing_leaving_transition_ids=[],
             existing_back_transition_ids=[],
             new_initial_state=False,
             old_origin_state_id=None,
             guard_name=None):
    """ """
    wfTool = api.portal.get_tool('portal_workflow')
    wf = wfTool.getWorkflowById(wf_id)
    if new_state_id in wf.states:
        return

    # ADD NEW STATE
    wf.states.addState(new_state_id)
    new_state = wf.states[new_state_id]
    new_state.setProperties(title=new_state_title, description='')
    # clone permissions
    clone_permissions(wf_id, permissions_cloned_state_id, new_state_id)
    # initial_state
    if new_initial_state:
        wf.initial_state = new_state_id

    # ADD NEW TRANSITIONS
    # leading_transition_id
    if leading_transition_id:
        wf.transitions.addTransition(leading_transition_id)
        transition = wf.transitions.get(leading_transition_id)
        if not guard_name:
            guard_name = 'may{0}{1}()'.format(leading_transition_id[0].upper(),
                                              leading_transition_id[1:])
        transition.setProperties(
            title=leading_transition_title,
            new_state_id=new_state_id,
            trigger_type=1,
            script_name='',
            actbox_name=leading_transition_id,
            actbox_url='',
            actbox_icon='%(portal_url)s/{0}.png'.format(leading_transition_id),
            actbox_category='workflow',
            props={'guard_expr': 'python:here.wfConditions().{0}'.format(guard_name)})
    # back_transition_id
    for back_transition_infos in back_transitions:
        back_transition_id = back_transition_infos['back_transition_id']
        back_transition_title = back_transition_infos['back_transition_title']
        back_from_state_id = back_transition_infos['back_from_state_id']
        if back_transition_id not in wf.transitions:
            wf.transitions.addTransition(back_transition_id)
        back_transition = wf.transitions.get(back_transition_id)
        back_transition.setProperties(
            title=back_transition_title,
            new_state_id=new_state_id,
            trigger_type=1,
            script_name='',
            actbox_name=back_transition_id,
            actbox_url='',
            actbox_icon='%(portal_url)s/{0}.png'.format(back_transition_id),
            actbox_category='workflow',
            props={'guard_expr': 'python:here.wfConditions().mayCorrect("{0}")'.format(new_state_id)})
        # add back_transition to back_from_state
        back_from_state = wf.states[back_from_state_id]
        back_from_state.transitions = back_from_state.transitions + (back_transition_id, )

    # leaving_transition_id
    if leaving_transition_id:
        wf.transitions.addTransition(leaving_transition_id)
        transition = wf.transitions.get(leaving_transition_id)
        guard_name = 'may{0}{1}'.format(leaving_transition_id[0].upper(), leaving_transition_id[1:])
        transition.setProperties(
            title=leaving_transition_id,
            new_state_id=leaving_to_state_id,
            trigger_type=1,
            script_name='',
            actbox_name=leaving_transition_id,
            actbox_url='',
            actbox_icon='%(portal_url)s/{0}.png'.format(leaving_transition_id),
            actbox_category='workflow',
            props={'guard_expr': 'python:here.wfConditions().{0}()'.format(guard_name)})

    # CONNECT STATES AND TRANSITIONS
    new_state.transitions = tuple([transition_id for transition_id in
                                  [leaving_transition_id] + existing_leaving_transition_ids
                                  if transition_id])

    # UPDATE connection between old origin state and new state
    # update also back transition between leaving_to_state_id
    if old_origin_state_id:
        old_origin_state = wf.states[old_origin_state_id]
        old_origin_state.transitions = [leading_transition_id]
        # remove transition to old_origin_state
        # and add this transition on the new created state
        leaving_to_state = wf.states[leaving_to_state_id]
        for transition_id in leaving_to_state.transitions:
            transition = wf.transitions[transition_id]
            if transition.new_state_id == old_origin_state_id:
                leaving_to_state_transition_ids = list(leaving_to_state.transitions)
                leaving_to_state_transition_ids.remove(transition_id)
                leaving_to_state.transitions = tuple(leaving_to_state_transition_ids)

    # existing_back_transitions
    for existing_back_transition_id in existing_back_transition_ids:
        existing_back_transition = wf.transitions.get(existing_back_transition_id)
        existing_back_transition.new_state_id = new_state_id


def removeState(wf, state_id, transition_id, back_transition_id):
    '''Made to handle removing a state from/to which only
       one transition is leaving/coming.'''
    # compute replacements transitions
    transition_id_repl = [tr for tr in wf.states[state_id].transitions
                          if not tr.startswith('backTo')][0]
    back_transition_id_repl = [tr for tr in wf.states[state_id].transitions
                               if tr.startswith('backTo')][0]
    # could be:
    # {'itemfreeze': 'itempublish', 'backToItemFrozen': 'backToPresented'}
    replacements = {transition_id: transition_id_repl,
                    back_transition_id: back_transition_id_repl}
    for state in wf.states.values():
        transitions = state.transitions
        for value, replacement in replacements.items():
            transitions = tuple(replace_in_list(transitions, value, replacement))
        state.transitions = transitions
    # Delete replaced transitions
    for tr in (transition_id, back_transition_id):
        if tr in wf.transitions:
            wf.transitions.deleteTransitions([tr])
    # Delete state 'itemfrozen'
    if state_id in wf.states:
        wf.states.deleteStates([state_id])


def clone_permissions(wf_id, base_state_id, new_state_id):
    ''' '''
    wfTool = api.portal.get_tool('portal_workflow')
    wf = wfTool.getWorkflowById(wf_id)
    base_state = wf.states[base_state_id]
    new_state = wf.states[new_state_id]
    for permission, roles in base_state.permission_roles.iteritems():
        # if roles is a list, it means it is acquired
        new_state.setPermission(permission, isinstance(roles, list) and 1 or 0, roles)


def _addDecidedState(new_state_id,
                     transition_id,
                     itemWorkflow,
                     base_state_id='accepted'):
    """Helper method for adding a decided state, base work will be done using the
       p_base_state (cloning permission, transition start/end points)."""
    wf = itemWorkflow

    # create new state
    wf.states.addState(new_state_id)

    # create transitions, for the 'back' transition, take the same as
    # when coming back from base_state_id
    wf.transitions.addTransition(transition_id)
    transition = wf.transitions[transition_id]
    transition.setProperties(
        title=transition_id,
        new_state_id=new_state_id, trigger_type=1, script_name='',
        actbox_name=transition_id, actbox_url='',
        actbox_icon='%(portal_url)s/{0}.png'.format(transition_id), actbox_category='workflow',
        props={'guard_expr': 'python:here.wfConditions().mayDecide()'})

    # use same transitions as state base_state_id
    back_transition_ids = wf.states[base_state_id].transitions
    # link state and transitions
    wf.states[new_state_id].setProperties(
        title=new_state_id, description='',
        transitions=back_transition_ids)

    # add state to possible transitions of same origin state as for base_state
    # get the transition leading to base_state then get the state it is going from
    tr_leading_to_base_state = [tr for tr in wf.transitions.values()
                                if tr.new_state_id == base_state_id][0].id
    # get the state, the transition 'delay' is going from
    origin_state_id = [state for state in wf.states.values()
                       if tr_leading_to_base_state in state.transitions][0].id
    wf.states[origin_state_id].transitions = \
        wf.states[origin_state_id].transitions + (transition_id, )

    # use same permissions as used by the base_state
    base_state = wf.states[base_state_id]
    new_state = wf.states[new_state_id]
    for permission, roles in base_state.permission_roles.iteritems():
        new_state.setPermission(permission, 0, roles)


def _addIsolatedState(new_state_id,
                      origin_state_id,
                      origin_transition_id,
                      origin_transition_guard_expr_name,
                      back_transition_id,
                      itemWorkflow,
                      back_transition_guard_expr_name='mayCorrect()',
                      base_state_id='accepted',
                      origin_transition_title=None,
                      origin_transition_icon=None,
                      back_transition_title=None,
                      back_transition_icon=None):
    """Add an isolated state with transitions go and back from/to another state."""
    wf = itemWorkflow
    # create new state
    wf.states.addState(new_state_id)
    new_state = wf.states[new_state_id]
    # use same permissions as used by the base_state_id state (default 'accepted')
    base_state = wf.states[base_state_id]
    for permission, roles in base_state.permission_roles.iteritems():
        new_state.setPermission(permission, 0, roles)

    # transitions
    # manage title if not given
    if origin_transition_title is None:
        origin_transition_title = origin_transition_id
    if back_transition_title is None:
        back_transition_title = back_transition_id
    # manage iconif not given
    if not origin_transition_icon:
        origin_transition_icon = origin_transition_id
    if not back_transition_icon:
        back_transition_icon = back_transition_id
    for transition_id, transition_title, destination_state_id, guard_expr_name, transition_icon in (
            (origin_transition_id, origin_transition_title, new_state_id,
             origin_transition_guard_expr_name, origin_transition_icon),
            (back_transition_id, back_transition_title, origin_state_id,
             back_transition_guard_expr_name, back_transition_icon)):
        if transition_id not in wf.transitions:
            wf.transitions.addTransition(transition_id)
            transition = wf.transitions[transition_id]
            transition.setProperties(
                title=transition_title,
                new_state_id=destination_state_id, trigger_type=1, script_name='',
                actbox_name=transition_id, actbox_url='',
                actbox_icon='%(portal_url)s/{0}.png'.format(transition_icon),
                actbox_category='workflow',
                props={'guard_expr': 'python:here.wfConditions().{0}'.format(guard_expr_name)})

    # link states and transitions
    # new_state
    new_state.setProperties(
        title=new_state_id, description='',
        transitions=[back_transition_id])
    # validate_state
    origin_state = wf.states[origin_state_id]
    origin_state.setProperties(
        title=origin_state.title, description=origin_state.description,
        transitions=origin_state.transitions + (origin_transition_id, ))
    return new_state


def get_base_item_validation_state():
    """The base item validation state is the state "validated" in the meetingitem_workflow.
       We use it from the base item WF because it manages the 'Add annex' permission correctly,
       the real final 'validated' state in itemWF is patched by '_apply_item_validation_levels'."""
    wfTool = api.portal.get_tool('portal_workflow')
    return wfTool.get('meetingitem_workflow').states['validated']


def _performWorkflowAdaptations(meetingConfig, logger=logger):
    '''This function applies workflow adaptations as specified by the p_meetingConfig.'''

    # Hereafter, adaptations are applied in some meaningful sequence:
    # adaptations that perform important structural changes like adding or
    # removing states and transitions are applied first; adaptations that work
    # only on role/permission mappings are applied at the end, so they apply on
    # a potentially modified set of states and transitions. Conflictual
    # combinations of adaptations exist, wrong combination of adaptations is
    # performed in meetingConfig.validate_workflowAdaptations.
    # If p_specificAdaptation is passed, just the relevant wfAdaptation is applied.
    wfAdaptations = meetingConfig.getWorkflowAdaptations()
    # make sure given wfAdaptations are in the right order
    # import MeetingConfig only here so we are sure that the 'wfAdaptations' attr
    # has been updated by subplugins if any
    ordered_wfAdaptations = get_vocab_values(meetingConfig, 'WorkflowAdaptations', sorted=False)
    wfAdaptations = list(wfAdaptations)
    wfAdaptations.sort(key=lambda x: ordered_wfAdaptations.index(x))

    itemWorkflow = meetingConfig.getItemWorkflow(True)
    meetingWorkflow = meetingConfig.getMeetingWorkflow(True)

    def _doWichValidationWithReturnedToProposingGroup(new_state_id,
                                                      base_state_id,
                                                      last_returned_state_id):
        """Helper method for adding a new state, base work will be done using the
           p_base_state_id (cloning permission, transition start/end points)."""
        wf = itemWorkflow

        # create new state
        wf.states.addState(new_state_id)

        # create transitions, between new state and last_returned_state_id
        # when coming back from base_state_id
        transition_id = 'backTo_%s' % last_returned_state_id
        leading_state_id = last_returned_state_id.replace('returned_to_proposing_group_', '')
        leading_state_title = safe_unicode(wf.states[leading_state_id].title)

        wf.transitions.addTransition(transition_id)
        transition = wf.transitions[transition_id]
        # manage translation of transition title when using validations
        if transition_id == "backTo_returned_to_proposing_group":
            # stored as utf-8
            transition_title = translate(
                'back_to_returned_to_proposing_group',
                domain="plone",
                context=meetingConfig.REQUEST).encode('utf-8')
        else:
            # stored as utf-8
            transition_title = translate(
                'back_to_returned_to_proposing_group_with_validation_state',
                domain="plone",
                mapping={"validation_state":
                    translate(leading_state_title,
                              domain='plone',
                              context=meetingConfig.REQUEST), },
                context=meetingConfig.REQUEST).encode('utf-8')
        image_url = '%(portal_url)s/{0}.png'.format(transition_id)
        transition.setProperties(
            title=transition_title,
            new_state_id=last_returned_state_id, trigger_type=1, script_name='',
            actbox_name=transition_id, actbox_url='',
            actbox_icon=image_url, actbox_category='workflow',
            props={'guard_expr': 'python:here.wfConditions().mayCorrect("%s")' % new_state_id})

        # use same transitions as state last_returned_state_id and add transition between
        # new state and last_returned_state_id except transitions start with backTo_returned_...
        back_transition_ids = tuple(
            [back_tr for back_tr in wf.states[last_returned_state_id].transitions
             if not back_tr.startswith('backTo_returned_')]) + (transition_id, )

        # link state and transitions
        wf.states[new_state_id].setProperties(
            title=new_state_id, description='',
            transitions=wf.states[new_state_id].transitions + back_transition_ids)

        # create transition between last_returned_state_id and new_state
        transition_id = 'goTo_%s' % (new_state_id)
        wf.transitions.addTransition(transition_id)
        transition = wf.transitions[transition_id]
        transition_title = meetingConfig.getItemWFValidationLevels(
            states=[base_state_id])['leading_transition_title']
        image_url = '%(portal_url)s/{0}.png'.format(transition_id)
        transition.setProperties(
            title=transition_title,
            new_state_id=new_state_id, trigger_type=1, script_name='',
            actbox_name=transition_id, actbox_url='',
            actbox_icon=image_url, actbox_category='workflow',
            props={
                'guard_expr': 'python:here.wfConditions()'
                '.mayProposeToNextValidationLevel(destinationState="{0}")'.format(
                    transition_id.replace('goTo_returned_to_proposing_group_', ''))})

        wf.states[last_returned_state_id].transitions = \
            wf.states[last_returned_state_id].transitions + (transition_id, )

        # use same permissions as used by the base_state
        base_state = wf.states[base_state_id]
        leading_state_id = new_state_id.replace('returned_to_proposing_group_', '')
        leading_state_title = safe_unicode(wf.states[leading_state_id].title)
        new_state_title = translate(
            'returned_to_proposing_group_with_validation_state',
            domain="plone",
            mapping={"validation_state":
                translate(leading_state_title,
                          domain='plone',
                          context=meetingConfig.REQUEST), },
            context=meetingConfig.REQUEST).encode('utf-8')
        new_state = wf.states[new_state_id]
        cloned_permissions = dict(base_state.permission_roles)
        new_state.permission_roles = cloned_permissions
        new_state.title = new_state_title

    def _apply_return_to_proposing_group(whichValidation=None):
        """Helper method to apply the 'return_to_proposing_group' or
           'return_to_proposing_group_with_last_validation' or
           'return_to_proposing_group_with_all_validations' wfAdaptation.
           whichValidation must in ('None', 'last', 'all')
        """
        if 'returned_to_proposing_group' not in itemWorkflow.states:
            itemWorkflow.states.addState('returned_to_proposing_group')
            new_state = getattr(itemWorkflow.states, 'returned_to_proposing_group')
            stateToClone = get_base_item_validation_state()
            # remove DeleteObjects permission
            cloned_permissions = dict(stateToClone.permission_roles)
            cloned_permissions[DeleteObjects] = ('Manager', )
            new_state.permission_roles = cloned_permissions

            # now create the necessary transitions : one to go to 'returned_to_proposing_group' state
            # and x to go back to relevant state depending on current meeting state
            # first, the transition 'return_to_proposing_group'
            itemWorkflow.transitions.addTransition('return_to_proposing_group')
            transition = itemWorkflow.transitions['return_to_proposing_group']
            transition.setProperties(
                title='return_to_proposing_group',
                new_state_id='returned_to_proposing_group', trigger_type=1, script_name='',
                actbox_name='return_to_proposing_group', actbox_url='',
                actbox_icon='%(portal_url)s/return_to_proposing_group.png', actbox_category='workflow',
                props={'guard_expr': 'python:here.wfConditions().mayReturnToProposingGroup()'})
            # Update connections between states and transitions and create new transitions
            newTransitionNames = []
            for stateName in RETURN_TO_PROPOSING_GROUP_FROM_ITEM_STATES:
                if stateName not in itemWorkflow.states:
                    continue
                # first specify that we can go to 'return_to_proposing_group' from this state
                currentTransitions = list(itemWorkflow.states[stateName].transitions)
                currentTransitions.append('return_to_proposing_group')
                itemWorkflow.states[stateName].transitions = tuple(currentTransitions)
                # then build a back transition name with given stateName
                transitionName = 'backTo_%s_from_returned_to_proposing_group' % stateName
                newTransitionNames.append(transitionName)
                itemWorkflow.transitions.addTransition(transitionName)
                transition = itemWorkflow.transitions[transitionName]
                # use a specific guard_expr 'mayBackToMeeting'
                transition.setProperties(
                    title='return_to_meeting',
                    new_state_id=stateName, trigger_type=1, script_name='',
                    actbox_name=transitionName, actbox_url='',
                    actbox_icon='%(portal_url)s/{0}.png'.format(transitionName), actbox_category='workflow',
                    props={'guard_expr': 'python:here.wfConditions().mayBackToMeeting("%s")' % transitionName})
                # now that we created back transitions, we can assign them to newState 'returned_to_proposing_group'
                # set properties for new 'returned_to_proposing_group' state
            new_state_title = translate(
                'returned_to_proposing_group',
                domain="plone",
                context=meetingConfig.REQUEST).encode('utf-8')
            new_state.setProperties(
                title=new_state_title, description='',
                transitions=newTransitionNames)

        # keep validation returned states
        validation_returned_states = _getValidationReturnedStates(meetingConfig)
        if whichValidation == 'last':
            validation_returned_states = (validation_returned_states[-1],)
        elif whichValidation is None:
            validation_returned_states = ()
        last_returned_state_id = 'returned_to_proposing_group'
        # as we may have groups without validation states, include the
        # 'returned_to_proposing_group' as state from which the item may be
        # sent back to the meeting, the mayBackToMeeting guard will handle this

        for validation_state in validation_returned_states:
            base_state_id = validation_state.replace('returned_to_proposing_group_', '')
            _doWichValidationWithReturnedToProposingGroup(new_state_id=validation_state,
                                                          base_state_id=base_state_id,
                                                          last_returned_state_id=last_returned_state_id)
            last_returned_state_id = validation_state

    def _apply_item_validation_levels(meetingConfig, itemWorkflow):
        """ """
        # build item validation levels based on MeetingConfig.itemWFValidationLevels values
        # build a list of dict with relevant informations so
        # we will be able to call addState.  Indeed we need "back transitions" ids
        # for example and we need to get this on the next validation level
        levels = []
        previous = None
        first = True
        item_validation_levels = list(meetingConfig.getItemWFValidationLevels(only_enabled=True))
        cfg_id = meetingConfig.getId(real_id=True)
        if not item_validation_levels:
            logger.info(WF_ITEM_VALIDATION_LEVELS_DISABLED % cfg_id)
            return

        # build thing reversed as workflows work with leading transitions
        # so we need to create transitions leading to a new state
        item_validation_levels.reverse()
        allowed_back_from_validated = get_allowed_back_shortcut_from(cfg_id, 'validated')
        for level in item_validation_levels:
            data = {}
            data['new_state_id'] = level['state']
            data['new_state_title'] = level['state_title']
            data['permissions_cloned_state_id'] = 'validated'
            data['leading_transition_id'] = level['leading_transition']
            data['leading_transition_title'] = level['leading_transition_title']
            data['back_transition'] = level['back_transition']
            data['back_transition_title'] = level['back_transition_title']
            data['guard_name'] = \
                'mayProposeToNextValidationLevel(destinationState="{0}")'.format(level['state'])
            # add transition from validated to every validation levels if allowed
            data['back_transitions'] = []
            if allowed_back_from_validated == '*' or \
               data['new_state_id'] in allowed_back_from_validated:
                data['back_transitions'].append(
                    {'back_transition_id': data['back_transition'],
                     'back_transition_title': data['back_transition_title'],
                     'back_from_state_id': 'validated'})
            if first:
                first = False
                data['leaving_to_state_id'] = 'validated'
                data['leaving_transition_id'] = 'validate'
            if previous:
                data['back_transitions'].append(
                    {'back_transition_id': data['back_transition'],
                     'back_transition_title': data['back_transition_title'],
                     'back_from_state_id': previous['new_state_id']})
                data['existing_leaving_transition_ids'] = ['validate', previous['leading_transition_id']]
                data['leaving_to_state_id'] = previous['new_state_id']
            previous = data
            levels.append(data)

        # make sure we have no leading transition for initial_state, it is the case
        # by default there is something because datagrid column is required...
        levels[-1]['leading_transition_id'] = ''
        levels[-1]['leading_transition_title'] = ''

        # last added level is the new initial state
        data['new_initial_state'] = True
        for level in levels:
            # remove unneeded 'back_transition/back_transition_title' keys, not used by addState
            level.pop('back_transition')
            level.pop('back_transition_title')
            addState(itemWorkflow.id, **level)

        # manage item_validation_shortcuts when every states and transitions are created
        if 'item_validation_shortcuts' in meetingConfig.getWorkflowAdaptations():
            # every transitions exist, we just need to add it to every item validation states
            back_shortcuts = {}
            # add back transitions
            for level in levels:
                back_shortcuts[level['new_state_id']] = []
                for back_shortcut_state in back_shortcuts:
                    allowed = get_allowed_back_shortcut_from(
                        meetingConfig.getId(real_id=True), back_shortcut_state)
                    if back_shortcut_state != level['new_state_id'] and \
                       (allowed == '*' or level['new_state_id'] in allowed):
                        # take last back_transition, the first is the one to validated if allowed
                        back_shortcuts[back_shortcut_state].append(
                            level['back_transitions'][-1]['back_transition_id'])
            levels.reverse()
            shortcuts = {}
            for level in levels:
                shortcuts[level['new_state_id']] = []
                for shortcut_state in shortcuts:
                    if shortcut_state != level['new_state_id']:
                        shortcuts[shortcut_state].append(
                            level['leading_transition_id'])
            # now update list of transitions leaving item validation states
            shortcuts = merge_dicts((shortcuts, back_shortcuts))
            for state_id, transition_ids in shortcuts.items():
                itemWorkflow.states[state_id].transitions = tuple(
                    set(itemWorkflow.states[state_id].transitions).union(transition_ids))

        # change permission for PloneMeeting: add annex for state "validated"
        # replace "Contributor" by "MeetingManager"
        # we use "validated" as base state this is why 'Add annex' permission
        # is given to ('Manager', 'Editor'), but finally we must change this
        # for the final 'validated' state
        validated = itemWorkflow.states["validated"]
        assert(validated.permission_roles[AddAnnex] == ('Manager', 'Editor'))
        validated.permission_roles[AddAnnex] = ('Manager', 'MeetingManager')

    tool = api.portal.get_tool('portal_plonemeeting')
    # first of all, manage MeetingConfig.itemWFValidationLevels
    _apply_item_validation_levels(meetingConfig, itemWorkflow)

    for wfAdaptation in wfAdaptations:
        # first try to call a performCustomWFAdaptations to see if it manages wfAdaptation
        # it could be a separated one or an overrided one
        applied = tool.adapted().performCustomWFAdaptations(
            meetingConfig, wfAdaptation, logger, itemWorkflow, meetingWorkflow)
        # double check if applied is True or False, we need that boolean
        if not isinstance(applied, bool):
            raise Exception('ToolPloneMeeting.performCustomWFAdaptations must return a boolean value!')
        # if performCustomWFAdaptations managed wfAdaptation, continue with next one
        if applied:
            logger.info(WF_APPLIED_CUSTOM % (wfAdaptation, meetingConfig.getId()))
            continue

        # "no_freeze" removes state 'frozen' in the meeting workflow and
        # corresponding state 'itemfrozen' in the item workflow.
        if wfAdaptation == 'no_freeze':
            # First, update the meeting workflow
            removeState(meetingWorkflow, 'frozen', 'freeze', 'backToFrozen')
            # Then, update the item workflow.
            removeState(itemWorkflow, 'itemfrozen', 'itemfreeze', 'backToItemFrozen')

        # "no_publication" removes state 'published' in the meeting workflow and
        # corresponding state 'itempublished' in the item workflow.
        if wfAdaptation == 'no_publication':
            # First, update the meeting workflow
            removeState(meetingWorkflow, 'published', 'publish', 'backToPublished')
            # Then, update the item workflow.
            removeState(itemWorkflow, 'itempublished', 'itempublish', 'backToItemPublished')

        # "itemdecided" adds state 'itemdecided' in the item workflow.
        if wfAdaptation == 'itemdecided':
            # Update the item workflow
            addState(
                wf_id=itemWorkflow.getId(),
                new_state_id='itemdecided',
                new_state_title='itemdecided',
                permissions_cloned_state_id='itemfrozen',
                leading_transition_id='itemdecide',
                leading_transition_title='itemdecide',
                back_transitions=[
                    {'back_transition_id': 'backToItemDecided',
                     'back_transition_title': 'backToItemDecided',
                     'back_from_state_id': 'accepted'}],
                existing_leaving_transition_ids=['accept', 'backToItemPublished'],
                leaving_to_state_id='accepted',
                guard_name='mayItemDecide()')
            # clean itemfrozen and accepted transitions
            itemWorkflow.states['accepted'].transitions = ['backToItemDecided']
            itemWorkflow.states['itempublished'].transitions = ['itemdecide', 'backToItemFrozen']

        # "no_decide" removes state 'decided' in the meeting workflow.
        if wfAdaptation == 'no_decide':
            # Update the meeting workflow
            removeState(meetingWorkflow, 'decided', 'decide', 'backToDecided')

        # "reviewers_take_back_validated_item" give the ability to reviewers to
        # take back an item that is validated.
        # This is managed in MeetingItem.MeetingItemWorkflowConditions.mayCorrect
        elif wfAdaptation == 'reviewers_take_back_validated_item':
            pass

        # "decide_item_when_back_to_meeting_from_returned_to_proposing_group" will
        # trigger some transitions defined in
        # config.ITEM_TRANSITION_WHEN_RETURNED_FROM_PROPOSING_GROUP_AFTER_CORRECTION
        # when an item that was returned_to_proposing_group
        # is send back to meeting if meeting is in a "decided" state
        elif wfAdaptation == 'decide_item_when_back_to_meeting_from_returned_to_proposing_group':
            pass

        # "only_creator_may_delete" grants the permission to delete items to
        # creators only (=role MeetingMember)(and also to God=Manager).
        # We will check states in which MeetingMember could delete and let only him
        # have the delete permission.  In states where MeetingMember could not delete,
        # nobody will be able to delete at all (except God Itself obviously)
        elif wfAdaptation == 'only_creator_may_delete':
            for state in itemWorkflow.states.values():
                if state.id != itemWorkflow.initial_state:
                    state.setPermission(DeleteObjects, 0, ['Manager'])

        # when an item is linked to a meeting, most of times, creators lose modify rights on it
        # with this, the item can be 'returned_to_proposing_group' when in a meeting then the creators
        # can modify it if necessary and send it back to the MeetingManagers when done
        elif wfAdaptation == 'return_to_proposing_group':
            _apply_return_to_proposing_group(whichValidation=None)

        # same as the "return_to_proposing_group" here above but the reviewer must validate item
        elif wfAdaptation == 'return_to_proposing_group_with_last_validation':
            _apply_return_to_proposing_group(whichValidation='last')

        # same as the "return_to_proposing_group" here above but the item must be validate by all hierarchical level
        elif wfAdaptation == 'return_to_proposing_group_with_all_validations':
            _apply_return_to_proposing_group(whichValidation='all')

        # "hide_decisions_when_under_writing" add state 'decisions_published' in the meeting workflow
        # between the 'decided' and the 'closed' states.  The idea is to hide the decisions to non
        # MeetingManagers when the meeting is 'decided' and to let everyone (that can see the item) access
        # the decision when the meeting state is 'decisions_published'.
        # To do so, we take this wfAdaptation into account in the MeetingItem.getDecision that hides it
        # for non MeetingManagers
        elif wfAdaptation == 'hide_decisions_when_under_writing':
            wf = meetingWorkflow
            # add state 'decision_published'
            if 'decisions_published' not in wf.states:
                wf.states.addState('decisions_published')
                # Create new transitions linking the new state to existing ones
                # ('decided' and 'closed').
                # add transitions 'publish_decisions' and 'backToDecisionsPublished'
                for tr in ('publish_decisions', 'backToDecisionsPublished'):
                    if tr not in wf.transitions:
                        wf.transitions.addTransition(tr)
                transition = wf.transitions['publish_decisions']
                transition.setProperties(
                    title='publish_decisions',
                    new_state_id='decisions_published', trigger_type=1, script_name='',
                    actbox_name='publish_decisions', actbox_url='',
                    actbox_icon='', actbox_category='workflow',
                    props={'guard_expr': 'python:here.wfConditions().mayPublishDecisions()'})
                transition = wf.transitions['backToDecisionsPublished']
                transition.setProperties(
                    title='backToDecisionsPublished',
                    new_state_id='decisions_published', trigger_type=1, script_name='',
                    actbox_name='backToDecisionsPublished', actbox_url='',
                    actbox_icon='', actbox_category='workflow',
                    props={'guard_expr': 'python:here.wfConditions().mayCorrect("decisions_published")'})
                # Update connections between states and transitions
                wf.states['decided'].setProperties(
                    title='decided', description='',
                    transitions=['backToFrozen', 'publish_decisions'])
                wf.states['decisions_published'].setProperties(
                    title='decisions_published', description='',
                    transitions=['backToDecided', 'close'])
                wf.states['closed'].setProperties(
                    title='closed', description='',
                    transitions=['backToDecisionsPublished', ])
                # Initialize permission->roles mapping for new state "decisions_published",
                # which is the same as state "frozen" (or "decided") in the previous setting.
                frozen = wf.states['frozen']
                decisions_published = wf.states['decisions_published']
                for permission, roles in frozen.permission_roles.iteritems():
                    decisions_published.setPermission(permission, 0, roles)
                # Transition "backToPublished" must be protected by a popup, like
                # any other "correct"-like transition.
                toConfirm = meetingConfig.getTransitionsToConfirm()
                if 'Meeting.backToDecisionsPublished' not in toConfirm:
                    toConfirm = list(toConfirm)
                    toConfirm.append('Meeting.backToDecisionsPublished')
                    meetingConfig.setTransitionsToConfirm(toConfirm)
                # State "decisions_published" must be selected in decisions DashboardCollections
                for collection in object_values(meetingConfig.searches.searches_decisions, 'DashboardCollection'):
                    for criterion in collection.query:
                        if criterion['i'] == 'review_state' and \
                           'decisions_published' not in criterion['v']:
                            updateCollectionCriterion(collection, criterion['i'],
                                                      tuple(criterion['v']) + ('decisions_published', ))

        # "waiting_advices/waiting_advices_from_last_validation_level"
        # add state 'xxx_waiting_advices' in the item workflow
        # it is a go/back state from the WAITING_ADVICES_FROM_STATES item list of states.
        # It is made to isolate an item in a state where it is no more editable but some advices may be given
        # if we have several 'xxx_waiting_advices' states added,
        # it is prefixed with originState1__or__originState2 like 'proposed__or__prevalidated_waiting_advices'
        elif wfAdaptation == 'waiting_advices':

            def _compute_from_transition_id(wf, from_state_id, new_state_id):
                from_transition_id = WAITING_ADVICES_FROM_TRANSITION_ID_PATTERN.format(
                    from_state_id)
                if from_transition_id in wf.transitions:
                    from_transition_id = WAITING_ADVICES_FROM_TO_TRANSITION_ID_PATTERN.format(
                        from_state_id, new_state_id)
                return from_transition_id

            wf = itemWorkflow
            # compute edit permissions existing on MeetingItem schema
            from Products.PloneMeeting.MeetingItem import MeetingItem
            edit_permissions = [ModifyPortalContent, DeleteObjects]
            for field in MeetingItem.schema.fields():
                # in some case we protect edit with "View" permission because
                # we manage access manually, ignore these edit permissions
                if field.write_permission and \
                   field.write_permission not in edit_permissions and \
                   field.write_permission not in (View, AccessContentsInformation):
                    edit_permissions.append(field.write_permission)
            new_state_id_pattern = '{0}_waiting_advices'
            # try to get meetingConfig id in WAITING_ADVICES_FROM_STATES
            # if not found, look for a "*" that is applied to every meetingConfigs
            # else nothing is done
            # for transition to 'xxx_waiting_advices', we need to know where we are coming from
            for infos in get_waiting_advices_infos(meetingConfig.getId(real_id=True)):
                # while using WFAs 'waiting_advices_from_before_last_val_level'
                # or 'waiting_advices_from_last_val_level', infos['from_states']/infos['back_states']
                # are ignored ad we use validation states
                from_states = []
                back_states = []
                if 'waiting_advices_from_every_val_levels' in wfAdaptations or \
                   'waiting_advices_from_before_last_val_level' in wfAdaptations or \
                   'waiting_advices_from_last_val_level' in wfAdaptations:
                    item_validation_states = meetingConfig.getItemWFValidationLevels(
                        data='state', only_enabled=True)
                    from_states = list(item_validation_states)
                    back_states = list(item_validation_states)
                else:
                    from_states = list(infos['from_states'])
                    back_states = list(infos['back_states'])
                if 'waiting_advices_adviser_may_validate' in wfAdaptations or \
                   infos.get('adviser_may_validate', False):
                    back_states.append('validated')
                # wipeout 'from_states' and 'back_states' to remove unexisting ones
                from_state_ids = [state for state in from_states if state in wf.states]
                back_state_ids = [state for state in back_states if state in wf.states]
                # if nothing left, continue
                if not from_state_ids or not back_state_ids:
                    continue
                new_state_id = infos.get('new_state_id', None) or \
                    new_state_id_pattern.format('__or__'.join(from_state_ids))
                if not new_state_id.endswith('_waiting_advices'):
                    raise Exception('Waiting advices "new_state_id" must end with "_waiting_advices" !')
                back_transition_ids = []
                if new_state_id not in wf.states:
                    wf.states.addState(new_state_id)
                    new_state = wf.states[new_state_id]
                    # Create new transitions to and from new_state
                    for from_state_id in from_state_ids:
                        from_transition_id = _compute_from_transition_id(
                            wf, from_state_id, new_state_id)
                        wf.transitions.addTransition(from_transition_id)
                        transition = wf.transitions[from_transition_id]
                        icon_name = 'wait_advices_from'
                        if infos.get('use_custom_icon', False):
                            icon_name = from_transition_id
                        tr_title = 'wait_advices_from'
                        if from_transition_id in infos.get('use_custom_transition_title_for', {}):
                            tr_title = infos['use_custom_transition_title_for'][from_transition_id]
                        transition.setProperties(
                            title=tr_title,
                            new_state_id=new_state_id, trigger_type=1, script_name='',
                            actbox_name=from_transition_id, actbox_url='',
                            actbox_icon='%(portal_url)s/{0}.png'.format(icon_name),
                            actbox_category='workflow',
                            props={
                                'guard_expr':
                                'python:here.wfConditions().mayWait_advices("{0}", "{1}")'.format(
                                    from_state_id, new_state_id)})
                        # update from_state
                        from_state = wf.states[from_state_id]
                        existing_transitions = from_state.transitions
                        from_state.transitions = existing_transitions + (from_transition_id, )

                    for back_state_id in back_state_ids:
                        defined_back_transition_ids = infos.get('defined_back_transition_ids', {})
                        if back_state_id in defined_back_transition_ids:
                            back_transition_id = defined_back_transition_ids[back_state_id]
                        else:
                            back_transition_id = 'backTo_{0}_from_waiting_advices'.format(back_state_id)
                        back_transition_ids.append(back_transition_id)
                        # we try to avoid creating too much back transitions
                        if back_transition_id not in wf.transitions:
                            wf.transitions.addTransition(back_transition_id)
                            transition = wf.transitions[back_transition_id]
                            back_transition_title = back_transition_id
                            icon_name = 'backTo_from_waiting_advices'
                            if back_state_id not in infos.get('use_custom_back_transition_title_for', ()):
                                # reuse the existing back transition for title and iconname
                                existing_back_transition_id = meetingConfig.getItemWFValidationLevels(
                                    states=[back_state_id], data='back_transition')
                                if existing_back_transition_id in wf.transitions:
                                    back_transition_title = wf.transitions[existing_back_transition_id].title
                                    icon_name = existing_back_transition_id
                            only_use_custom_back_transition_icon_for = \
                                infos.get('only_use_custom_back_transition_icon_for', ())
                            if not only_use_custom_back_transition_icon_for or \
                               back_state_id in only_use_custom_back_transition_icon_for:
                                icon_name = back_transition_id
                            transition.setProperties(
                                title=back_transition_title,
                                new_state_id=back_state_id, trigger_type=1, script_name='',
                                actbox_name=back_transition_id, actbox_url='',
                                actbox_icon='%(portal_url)s/{0}.png'.format(icon_name),
                                actbox_category='workflow',
                                props={'guard_expr': 'python:here.wfConditions().mayCorrect("%s")'
                                       % back_state_id})

                    # Update connections between states and transitions
                    new_state_title = new_state_id
                    if not infos.get('use_custom_state_title', True):
                        new_state_title = 'waiting_advices'
                    new_state.setProperties(title=new_state_title,
                                            description='',
                                            transitions=back_transition_ids)

                    # Initialize permission->roles mapping for new state "to_transition",
                    # which is the same as state 'validated'
                    perm_cloned_state = wf.states[infos.get('perm_cloned_state', 'validated')]
                    for permission, roles in perm_cloned_state.permission_roles.iteritems():
                        if infos.get('remove_modify_access', True) and permission in edit_permissions:
                            # remove every roles but 'Manager', 'MeetingManager',
                            # 'MeetingBudgetImpactEditor' and 'MeetingInternalNotesEditor'
                            # the intersection takes care of keeping the relevant roles
                            edit_roles = set(roles).intersection(
                                set(('Manager',
                                     'MeetingManager',
                                     'MeetingBudgetImpactEditor',
                                     'MeetingInternalNotesEditor')))
                            new_state.setPermission(permission, 0, edit_roles)
                        else:
                            new_state.setPermission(permission, 0, roles)

        # "postpone_next_meeting" add state 'postponed_next_meeting' in the item workflow
        # additionnaly, when an item is set to this state, it will be duplicated and validated
        # for a next meeting thru the doPostpone_next_meeting method
        elif wfAdaptation == 'postpone_next_meeting':
            _addDecidedState(new_state_id='postponed_next_meeting',
                             transition_id='postpone_next_meeting',
                             itemWorkflow=itemWorkflow)

        # "mark_not_applicable" add state 'marked_not_applicable' in the item workflow
        elif wfAdaptation == 'mark_not_applicable':
            _addDecidedState(new_state_id='marked_not_applicable',
                             transition_id='mark_not_applicable',
                             itemWorkflow=itemWorkflow)

        # "delayed" add state 'delayed' in the item workflow
        elif wfAdaptation == 'delayed':
            _addDecidedState(new_state_id='delayed',
                             transition_id='delay',
                             itemWorkflow=itemWorkflow)

        # "removed" and "removed_and_duplicated" add state 'removed' in the item workflow
        elif wfAdaptation in ('removed', 'removed_and_duplicated'):
            _addDecidedState(new_state_id='removed',
                             transition_id='remove',
                             itemWorkflow=itemWorkflow)

        # "refused" add state 'refused' in the item workflow
        elif wfAdaptation == 'refused':
            _addDecidedState(new_state_id='refused',
                             transition_id='refuse',
                             itemWorkflow=itemWorkflow)

        # "accepted_but_modified" add state 'accepted_but_modified' in the item workflow
        elif wfAdaptation == 'accepted_but_modified':
            _addDecidedState(new_state_id='accepted_but_modified',
                             transition_id='accept_but_modify',
                             itemWorkflow=itemWorkflow)

        # "pre_accepted" add state 'pre_accepted'
        # from 'itemfrozen' in the item WF
        elif wfAdaptation == 'pre_accepted':
            # add state from itemfrozen? itempublished? presented? ...
            # same origin as mandatory transition 'accept'
            origin_state_id = [state.id for state in itemWorkflow.states.values()
                               if 'accept' in state.transitions][0]
            back_transition_id = [tr for tr in itemWorkflow.states['accepted'].transitions
                                  if tr.startswith('backTo')][0]
            # we use the origin_state_id as base_state_id because
            # MeetingManager must be able to edit the item
            new_state = _addIsolatedState(
                new_state_id='pre_accepted',
                origin_state_id=origin_state_id,
                origin_transition_id='pre_accept',
                origin_transition_guard_expr_name='mayDecide()',
                back_transition_guard_expr_name="mayCorrect('%s')" % origin_state_id,
                back_transition_id=back_transition_id,
                itemWorkflow=itemWorkflow,
                base_state_id=origin_state_id, )
            # ... then add output transitions to 'accepted' and 'accepted_but_modified'
            out_transitions = new_state.transitions
            out_transitions += ('accept', )
            if 'accepted_but_modified' in wfAdaptations:
                out_transitions += ('accept_but_modify', )
            new_state.transitions = out_transitions

        # "accepted_out_of_meeting" add state 'accepted_out_of_meeting'
        # from 'validated' in the item WF
        elif wfAdaptation in ['accepted_out_of_meeting',
                              'accepted_out_of_meeting_and_duplicated']:
            _addIsolatedState(
                new_state_id='accepted_out_of_meeting',
                origin_state_id='validated',
                origin_transition_id='accept_out_of_meeting',
                origin_transition_guard_expr_name='mayAccept_out_of_meeting()',
                back_transition_guard_expr_name="mayCorrect('validated')",
                back_transition_id='backToValidatedFromAcceptedOutOfMeeting',
                itemWorkflow=itemWorkflow)

        # "accepted_out_of_meeting_emergency" add state 'accepted_out_of_meeting_emergency'
        # from 'validated' in the item WF
        elif wfAdaptation in ['accepted_out_of_meeting_emergency',
                              'accepted_out_of_meeting_emergency_and_duplicated']:
            _addIsolatedState(
                new_state_id='accepted_out_of_meeting_emergency',
                origin_state_id='validated',
                origin_transition_id='accept_out_of_meeting_emergency',
                origin_transition_guard_expr_name='mayAccept_out_of_meeting_emergency()',
                back_transition_guard_expr_name="mayCorrect('validated')",
                back_transition_id='backToValidatedFromAcceptedOutOfMeetingEmergency',
                itemWorkflow=itemWorkflow)

        # "transfered" add state 'transfered' from 'validated' in the item WF
        elif wfAdaptation in ['transfered', 'transfered_and_duplicated']:
            _addIsolatedState(
                new_state_id='transfered',
                origin_state_id='validated',
                origin_transition_id='transfer',
                origin_transition_guard_expr_name='mayTransfer()',
                back_transition_guard_expr_name="mayCorrect('validated')",
                back_transition_id='backToValidatedFromTransfered',
                itemWorkflow=itemWorkflow)

        # "presented_item_back_to_XXX" allows the MeetingManagers to send a presented
        # item directly back to "XXX" item validation state in addition to back to "validated"
        elif wfAdaptation.startswith('presented_item_back_to_'):
            presented = itemWorkflow.states.presented
            item_validation_state = wfAdaptation.replace('presented_item_back_to_', '')
            # find back transition that leads to item_validation_state
            validation_state_infos = meetingConfig.getItemWFValidationLevels(
                states=[item_validation_state])
            back_transition = validation_state_infos['back_transition']
            presented.transitions = presented.transitions + (back_transition, )

        elif wfAdaptation == MEETING_REMOVE_MOG_WFA:
            for state in meetingWorkflow.states.values():
                for permission in state.permission_roles:
                    removePermission(state, permission, "MeetingObserverGlobal")

        logger.info(WF_APPLIED % (wfAdaptation, meetingConfig.getId()))


def _performAdviceWorkflowAdaptations(logger=logger):
    '''This function applies advice related workflow adaptations.'''
    tool = api.portal.get_tool('portal_plonemeeting')
    # save patched wf_ids as several org uids may use same patched workflow
    patched_wf_ids = []
    for org_uid, extra_adviser_infos in tool.adapted().get_extra_adviser_infos().items():
        portal_type = extra_adviser_infos['portal_type']
        base_wf = extra_adviser_infos['base_wf']
        if base_wf in patched_wf_ids:
            continue
        patched_wf_ids.append(base_wf)
        advice_wf_id = '{0}__{1}'.format(portal_type, base_wf)
        wfAdaptations = extra_adviser_infos['wf_adaptations']
        for wfAdaptation in wfAdaptations:
            # first try to call a performCustomAdviceWFAdaptations
            applied = tool.adapted().performCustomAdviceWFAdaptations(
                tool, wfAdaptation, logger, advice_wf_id)
            # double check if applied is True or False, we need that boolean
            if not isinstance(applied, bool):
                raise Exception(
                    'ToolPloneMeeting.performCustomAdviceWFAdaptations must '
                    'return a boolean value!')
            # if performCustomAdviceWFAdaptations managed wfAdaptation,
            # continue with next one
            if applied:
                logger.info(ADVICE_WF_APPLIED_CUSTOM % (wfAdaptation, advice_wf_id))
                continue
            else:
                # nothing done by default for now
                raise Exception('ToolPloneMeeting.performCustomAdviceWFAdaptations '
                                'did not handle WFAdaptation %s!' % wfAdaptation)


def _getValidationReturnedStates(cfg):
    ''' '''
    res = []
    states = cfg.getItemWFValidationLevels(data='state', only_enabled=True)[1:]
    for state in states:
        res.append("returned_to_proposing_group_%s" % state)
    return res



# -*- coding: utf-8 -*-
#
# GNU General Public License (GPL)
#


from collective.behavior.talcondition.behavior import TALCondition
from Products.PloneMeeting.utils import _base_extra_expr_ctx


class PMTALCondition(TALCondition):
    """Override TALCondition behavior to add our own TAL expression extra context."""

    def complete_extra_expr_ctx(self, extra_expr_ctx):
        """Return extra_expr_ctx, this is made to be overrided."""
        extra_context = super(PMTALCondition, self).complete_extra_expr_ctx(extra_expr_ctx)
        context = extra_context.get('context', self.context)
        base_extra_expr_ctx = _base_extra_expr_ctx(context)
        extra_context.update(base_extra_expr_ctx)
        return extra_context



# -*- coding: utf-8 -*-
#
# GNU General Public License (GPL)
#

from AccessControl import Unauthorized
from imio.helpers.content import get_vocab
from io import BytesIO
from plone.directives import form
from Products.CMFPlone import PloneMessageFactory as PMF
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.interfaces import IRedirect
from Products.PloneMeeting.widgets.pm_checkbox import PMCheckBoxFieldWidget
from PyPDF2 import PdfFileReader
from PyPDF2 import PdfFileWriter
from z3c.form import button
from z3c.form import field
from z3c.form import form as z3c_form
from zope import schema


class IItemExportPDF(form.Schema):
    """ """

    elements = schema.List(
        title=_(u"Elements to export in PDF"),
        description=_(u""),
        required=False,
        value_type=schema.Choice(
            vocabulary=u"Products.PloneMeeting.vocabularies.item_export_pdf_elements_vocabulary"),
    )


class ItemExportPDFForm(z3c_form.Form):
    """ """
    fields = field.Fields(IItemExportPDF)
    fields["elements"].widgetFactory = PMCheckBoxFieldWidget

    ignoreContext = True  # don't use context to get widget data

    label = PMF(u"Export PDF")
    description = _('export_pdf_descr')
    _finished = False

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @button.buttonAndHandler(_('Apply'), name='apply_export_pdf')
    def handleApply(self, action):
        self._check_auth()
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self._check_data(data)
        return self._do_export_pdf(data)

    def updateWidgets(self):
        super(ItemExportPDFForm, self).updateWidgets()
        self.widgets['elements'].sortable = True

    def _check_data(self, data):
        """Make sure elements are selectable.
           As some values are disabled in the UI, a user could try
           to surround this, raise Unauthorized in this case."""
        selectable = [term.token for term in self.widgets['elements'].terms
                      if not getattr(term, 'disabled', False)]
        for elt_id in data['elements']:
            if elt_id not in selectable:
                raise Unauthorized

    def _do_export_pdf(self, data):
        # pod templates
        view = self.context.restrictedTraverse('@@document-generation')
        pod_template_uids = [
            term.token for term in get_vocab(
                self.context,
                'Products.PloneMeeting.vocabularies.'
                'generable_pdf_documents_vocabulary')._terms
            if term.token in data['elements']]
        content = {template_uid: view(template_uid=template_uid, output_format='pdf')
                   for template_uid in pod_template_uids}
        # annexes
        content.update({annex_id: self.context.get(annex_id).file.data
                        for annex_id in data['elements'] if annex_id not in pod_template_uids})
        # create unique PDF file
        output_writer = PdfFileWriter()
        for element_id in data['elements']:
            output_writer.appendPagesFromReader(
                PdfFileReader(BytesIO(content[element_id]), strict=False))
        pdf_file_content = BytesIO()
        output_writer.write(pdf_file_content)
        self.request.set('pdf_file_content', pdf_file_content)
        return pdf_file_content

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self._finished = True

    def update(self):
        self._check_auth()
        super(ItemExportPDFForm, self).update()
        # after calling parent's update, self.actions are available
        self.actions.get('cancel').addClass('standalone')

    def _check_auth(self):
        """Raise Unauthorized if current user may not export to PDF."""
        if not self.context.show_export_pdf_action():
            raise Unauthorized

    def render(self):
        if 'pdf_file_content' in self.request:
            filename = "export_pdf_%s" % self.context.getId()
            self.request.response.setHeader('Content-Type', 'application/pdf')
            self.request.response.setHeader('Content-disposition', 'attachment;filename=%s.pdf' % filename)
            pdf_file_content = self.request['pdf_file_content']
            pdf_file_content.seek(0)
            return pdf_file_content.read()

        if self._finished:
            IRedirect(self.request).redirect(self.context.absolute_url())
            return ""
        return super(ItemExportPDFForm, self).render()
